import errno
import os
import socket
import tempfile
from itertools import cycle
from unittest.mock import MagicMock, Mock, patch

import pytest
from billiard.pool import ApplyResult
from kombu.asynchronous import Hub

import t.skip
from celery.app.defaults import DEFAULTS
from celery.concurrency.asynpool import iterate_file_descriptors_safely
from celery.utils.collections import AttributeDict
from celery.utils.functional import noop
from celery.utils.objects import Bunch

try:
    from celery.concurrency import asynpool
    from celery.concurrency import prefork as mp
except ImportError:

    class _mp:
        RUN = 0x1

        class TaskPool:
            _pool = Mock()

            def __init__(self, *args, **kwargs):
                pass

            def start(self):
                pass

            def stop(self):
                pass

            def apply_async(self, *args, **kwargs):
                pass
    mp = _mp()
    asynpool = None


class MockResult:

    def __init__(self, value, pid):
        self.value = value
        self.pid = pid

    def worker_pids(self):
        return [self.pid]

    def get(self):
        return self.value


@patch('celery.platforms.set_mp_process_title')
class test_process_initializer:

    @staticmethod
    def Loader(*args, **kwargs):
        loader = Mock(*args, **kwargs)
        loader.conf = {}
        loader.override_backends = {}
        return loader

    @patch('celery.platforms.signals')
    def test_process_initializer(self, _signals, set_mp_process_title, restore_logging):
        from celery import signals
        from celery._state import _tls
        from celery.concurrency.prefork import WORKER_SIGIGNORE, WORKER_SIGRESET, process_initializer
        on_worker_process_init = Mock()
        signals.worker_process_init.connect(on_worker_process_init)

        with self.Celery(loader=self.Loader) as app:
            app.conf = AttributeDict(DEFAULTS)
            process_initializer(app, 'awesome.worker.com')
            _signals.ignore.assert_any_call(*WORKER_SIGIGNORE)
            _signals.reset.assert_any_call(*WORKER_SIGRESET)
            assert app.loader.init_worker.call_count
            on_worker_process_init.assert_called()
            assert _tls.current_app is app
            set_mp_process_title.assert_called_with(
                'celeryd', hostname='awesome.worker.com',
            )

            with patch('celery.app.trace.setup_worker_optimizations') as S:
                os.environ['FORKED_BY_MULTIPROCESSING'] = '1'
                try:
                    process_initializer(app, 'luke.worker.com')
                    S.assert_called_with(app, 'luke.worker.com')
                finally:
                    os.environ.pop('FORKED_BY_MULTIPROCESSING', None)

            os.environ['CELERY_LOG_FILE'] = 'worker%I.log'
            app.log.setup = Mock(name='log_setup')
            try:
                process_initializer(app, 'luke.worker.com')
            finally:
                os.environ.pop('CELERY_LOG_FILE', None)

    @patch('celery.platforms.set_pdeathsig')
    def test_pdeath_sig(self, _set_pdeathsig, set_mp_process_title, restore_logging):
        from celery import signals
        on_worker_process_init = Mock()
        signals.worker_process_init.connect(on_worker_process_init)
        from celery.concurrency.prefork import process_initializer

        with self.Celery(loader=self.Loader) as app:
            app.conf = AttributeDict(DEFAULTS)
            process_initializer(app, 'awesome.worker.com')
        _set_pdeathsig.assert_called_once_with('SIGKILL')


class test_process_destructor:

    @patch('celery.concurrency.prefork.signals')
    def test_process_destructor(self, signals):
        mp.process_destructor(13, -3)
        signals.worker_process_shutdown.send.assert_called_with(
            sender=None, pid=13, exitcode=-3,
        )


class MockPool:
    started = False
    closed = False
    joined = False
    terminated = False
    _state = None

    def __init__(self, *args, **kwargs):
        self.started = True
        self._timeout_handler = Mock()
        self._result_handler = Mock()
        self.maintain_pool = Mock()
        self._state = mp.RUN
        self._processes = kwargs.get('processes')
        self._proc_alive_timeout = kwargs.get('proc_alive_timeout')
        self._pool = [Bunch(pid=i, inqW_fd=1, outqR_fd=2)
                      for i in range(self._processes)]
        self._current_proc = cycle(range(self._processes))

    def close(self):
        self.closed = True
        self._state = 'CLOSE'

    def join(self):
        self.joined = True

    def terminate(self):
        self.terminated = True

    def terminate_job(self, *args, **kwargs):
        pass

    def restart(self, *args, **kwargs):
        pass

    def handle_result_event(self, *args, **kwargs):
        pass

    def flush(self):
        pass

    def grow(self, n=1):
        self._processes += n

    def shrink(self, n=1):
        self._processes -= n

    def apply_async(self, *args, **kwargs):
        pass

    def register_with_event_loop(self, loop):
        pass


class ExeMockPool(MockPool):

    def apply_async(self, target, args=(), kwargs={}, callback=noop):
        from threading import Timer
        res = target(*args, **kwargs)
        Timer(0.1, callback, (res,)).start()
        return MockResult(res, next(self._current_proc))


class TaskPool(mp.TaskPool):
    Pool = BlockingPool = MockPool


class ExeMockTaskPool(mp.TaskPool):
    Pool = BlockingPool = ExeMockPool


@t.skip.if_win32
class test_AsynPool:

    def setup_method(self):
        pytest.importorskip('multiprocessing')

    def test_gen_not_started(self):

        def gen():
            yield 1
            assert not asynpool.gen_not_started(g)
            yield 2
        g = gen()
        assert asynpool.gen_not_started(g)
        next(g)
        assert not asynpool.gen_not_started(g)
        list(g)
        assert not asynpool.gen_not_started(g)

        def gen2():
            yield 1
            raise RuntimeError('generator error')
        g = gen2()
        assert asynpool.gen_not_started(g)
        next(g)
        assert not asynpool.gen_not_started(g)
        with pytest.raises(RuntimeError):
            next(g)
        assert not asynpool.gen_not_started(g)

    @patch('select.select', create=True)
    def test_select(self, __select):
        ebadf = socket.error()
        ebadf.errno = errno.EBADF
        with patch('select.poll', create=True) as poller:
            poll = poller.return_value = Mock(name='poll.poll')
            poll.return_value = {3}, set(), 0
            assert asynpool._select({3}, poll=poll) == ({3}, set(), 0)

            poll.return_value = {3}, set(), 0
            assert asynpool._select({3}, None, {3}, poll=poll) == (
                {3}, set(), 0,
            )

            eintr = socket.error()
            eintr.errno = errno.EINTR
            poll.side_effect = eintr

            readers = {3}
            assert asynpool._select(readers, poll=poll) == (set(), set(), 1)
            assert 3 in readers

        with patch('select.poll', create=True) as poller:
            poll = poller.return_value = Mock(name='poll.poll')
            poll.side_effect = ebadf
            with patch('select.select') as selcheck:
                selcheck.side_effect = ebadf
                readers = {3}
                assert asynpool._select(readers, poll=poll) == (
                    set(), set(), 1,
                )
                assert 3 not in readers

        with patch('select.poll', create=True) as poller:
            poll = poller.return_value = Mock(name='poll.poll')
            poll.side_effect = MemoryError()
            with pytest.raises(MemoryError):
                asynpool._select({1}, poll=poll)

        with patch('select.poll', create=True) as poller:
            poll = poller.return_value = Mock(name='poll.poll')
            with patch('select.select') as selcheck:

                def se(*args):
                    selcheck.side_effect = MemoryError()
                    raise ebadf
                poll.side_effect = se
                with pytest.raises(MemoryError):
                    asynpool._select({3}, poll=poll)

        with patch('select.poll', create=True) as poller:
            poll = poller.return_value = Mock(name='poll.poll')
            with patch('select.select') as selcheck:

                def se2(*args):
                    selcheck.side_effect = socket.error()
                    selcheck.side_effect.errno = 1321
                    raise ebadf
                poll.side_effect = se2
                with pytest.raises(socket.error):
                    asynpool._select({3}, poll=poll)

        with patch('select.poll', create=True) as poller:
            poll = poller.return_value = Mock(name='poll.poll')

            poll.side_effect = socket.error()
            poll.side_effect.errno = 34134
            with pytest.raises(socket.error):
                asynpool._select({3}, poll=poll)

    def test_select_unpatched(self):
        with tempfile.TemporaryFile('w') as f:
            _, writeable, _ = asynpool._select(writers={f, }, err={f, })
            assert f.fileno() in writeable

        with tempfile.TemporaryFile('r') as f:
            readable, _, _ = asynpool._select(readers={f, }, err={f, })
            assert f.fileno() in readable

    def test_promise(self):
        fun = Mock()
        x = asynpool.promise(fun, (1,), {'foo': 1})
        x()
        assert x.ready
        fun.assert_called_with(1, foo=1)

    def test_Worker(self):
        w = asynpool.Worker(Mock(), Mock())
        w.on_loop_start(1234)
        w.outq.put.assert_called_with((asynpool.WORKER_UP, (1234,)))

    def test_iterate_file_descriptors_safely_source_data_list(self):
        # Given: a list of integers that could be file descriptors
        fd_iter = [1, 2, 3, 4, 5]

        # Given: a mock hub method that does nothing to call
        def _fake_hub(*args, **kwargs):
            raise OSError

        # When Calling the helper to iterate_file_descriptors_safely
        iterate_file_descriptors_safely(
            fd_iter, fd_iter, _fake_hub,
            "arg1", "arg2", kw1="kw1", kw2="kw2",
        )

        # Then: all items were removed from the managed data source
        assert fd_iter == [], "Expected all items removed from managed list"

    def test_iterate_file_descriptors_safely_source_data_set(self):
        # Given: a list of integers that could be file descriptors
        fd_iter = {1, 2, 3, 4, 5}

        # Given: a mock hub method that does nothing to call
        def _fake_hub(*args, **kwargs):
            raise OSError

        # When Calling the helper to iterate_file_descriptors_safely
        iterate_file_descriptors_safely(
            fd_iter, fd_iter, _fake_hub,
            "arg1", "arg2", kw1="kw1", kw2="kw2",
        )

        # Then: all items were removed from the managed data source
        assert fd_iter == set(), "Expected all items removed from managed set"

    def test_iterate_file_descriptors_safely_source_data_dict(self):
        # Given: a list of integers that could be file descriptors
        fd_iter = {1: 1, 2: 2, 3: 3, 4: 4, 5: 5}

        # Given: a mock hub method that does nothing to call
        def _fake_hub(*args, **kwargs):
            raise OSError

        # When Calling the helper to iterate_file_descriptors_safely
        iterate_file_descriptors_safely(
            fd_iter, fd_iter, _fake_hub,
            "arg1", "arg2", kw1="kw1", kw2="kw2",
        )

        # Then: all items were removed from the managed data source
        assert fd_iter == {}, "Expected all items removed from managed dict"

    def _get_hub(self):
        hub = Hub()
        hub.readers = {}
        hub.writers = {}
        hub.timer = Mock(name='hub.timer')
        hub.timer._queue = [Mock()]
        hub.fire_timers = Mock(name='hub.fire_timers')
        hub.fire_timers.return_value = 1.7
        hub.poller = Mock(name='hub.poller')
        hub.close = Mock(name='hub.close()')
        return hub

    @t.skip.if_pypy
    def test_schedule_writes_hub_remove_writer_ready_fd_not_in_all_inqueues(self):
        pool = asynpool.AsynPool(threads=False)
        hub = self._get_hub()

        writer = Mock(name='writer')
        reader = Mock(name='reader')

        # add 2 fake fds with the same id
        hub.add_reader(6, reader, 6)
        hub.add_writer(6, writer, 6)
        pool._all_inqueues.clear()
        pool._create_write_handlers(hub)

        # check schedule_writes write fds remove not remove the reader one from the hub.
        hub.consolidate_callback(ready_fds=[6])
        assert 6 in hub.readers
        assert 6 not in hub.writers

    @t.skip.if_pypy
    def test_schedule_writes_hub_remove_writers_from_active_writers_when_get_index_error(self):
        pool = asynpool.AsynPool(threads=False)
        hub = self._get_hub()

        writer = Mock(name='writer')
        reader = Mock(name='reader')

        # add 3 fake fds with the same id to reader and writer
        hub.add_reader(6, reader, 6)
        hub.add_reader(8, reader, 8)
        hub.add_reader(9, reader, 9)
        hub.add_writer(6, writer, 6)
        hub.add_writer(8, writer, 8)
        hub.add_writer(9, writer, 9)

        # add fake fd to pool _all_inqueues to make sure we try to read from outbound_buffer
        # set active_writes to 6 to make sure we remove all write fds except 6
        pool._active_writes = {6}
        pool._all_inqueues = {2, 6, 8, 9}

        pool._create_write_handlers(hub)

        # clear outbound_buffer to get IndexError when trying to pop any message
        # in this case all active_writers fds will be removed from the hub
        pool.outbound_buffer.clear()

        hub.consolidate_callback(ready_fds=[2])
        if {6, 8, 9} <= hub.readers.keys() and not {8, 9} <= hub.writers.keys():
            assert True
        else:
            assert False

        assert 6 in hub.writers

    @t.skip.if_pypy
    def test_schedule_writes_hub_remove_fd_only_from_writers_when_write_job_is_done(self):
        pool = asynpool.AsynPool(threads=False)
        hub = self._get_hub()

        writer = Mock(name='writer')
        reader = Mock(name='reader')

        # add one writer and one reader with the same fd
        hub.add_writer(2, writer, 2)
        hub.add_reader(2, reader, 2)
        assert 2 in hub.writers

        # For test purposes to reach _write_job in schedule writes
        pool._all_inqueues = {2}
        worker = Mock("worker")
        # this lambda need to return a number higher than 4
        # to pass the while loop in _write_job function and to reach the hub.remove_writer
        worker.send_job_offset = lambda header, HW: 5

        pool._fileno_to_inq[2] = worker
        pool._create_write_handlers(hub)

        result = ApplyResult({}, lambda x: True)
        result._payload = [None, None, -1]
        pool.outbound_buffer.appendleft(result)

        hub.consolidate_callback(ready_fds=[2])
        assert 2 not in hub.writers
        assert 2 in hub.readers

    @t.skip.if_pypy
    def test_register_with_event_loop__no_on_tick_dupes(self):
        """Ensure AsynPool's register_with_event_loop only registers
        on_poll_start in the event loop the first time it's called. This
        prevents a leak when the Consumer is restarted.
        """
        pool = asynpool.AsynPool(threads=False)
        hub = Mock(name='hub')
        pool.register_with_event_loop(hub)
        pool.register_with_event_loop(hub)
        hub.on_tick.add.assert_called_once()

    @t.skip.if_pypy
    @patch('billiard.pool.Pool._create_worker_process')
    def test_before_create_process_signal(self, create_process):
        from celery import signals
        on_worker_before_create_process = Mock()
        signals.worker_before_create_process.connect(on_worker_before_create_process)
        pool = asynpool.AsynPool(processes=1, threads=False)
        create_process.assert_called_once_with(0)
        on_worker_before_create_process.assert_any_call(
            signal=signals.worker_before_create_process,
            sender=pool,
        )

    def test_untrack_child_process_without_sentinel_poll(self):
        """_untrack_child_process must not raise when proc lacks _sentinel_poll.

        Race condition during cold shutdown can cause _untrack_child_process to
        be called with a process that never had _sentinel_poll set or had it
        cleared. Use getattr for safe access.
        """
        pytest.importorskip('multiprocessing')
        pool = asynpool.AsynPool(processes=1, threads=False)
        hub = Mock(name='hub')
        proc = object()  # No _sentinel_poll attribute
        pool._untrack_child_process(proc, hub)  # Should not raise AttributeError
        hub.remove.assert_not_called()

    def test_untrack_child_process_with_sentinel_poll(self):
        """_untrack_child_process cleans up when proc has _sentinel_poll set."""
        pytest.importorskip('multiprocessing')
        pool = asynpool.AsynPool(processes=1, threads=False)
        hub = Mock(name='hub')
        fd = os.open(os.devnull, os.O_RDONLY)
        proc = Mock(_sentinel_poll=fd)
        pool._untrack_child_process(proc, hub)
        hub.remove.assert_called_once_with(fd)
        assert proc._sentinel_poll is None

    @t.skip.if_pypy
    def test_flush_no_synack_discards_unaccepted_jobs(self):
        """flush() should discard unaccepted jobs when synack is disabled.

        Previously, flush() only handled the synack case. Without synack,
        unaccepted jobs were never cleaned from the cache, leading to stale
        entries.
        """
        pool = asynpool.AsynPool(processes=1, synack=False, threads=False)
        pool._state = asynpool.RUN
        pool.maintain_pool = Mock(name='maintain_pool')

        job1 = Mock(name='job1')
        job1._accepted = False
        job1._writer.return_value = None
        job2 = Mock(name='job2')
        job2._accepted = True
        job2._writer.return_value = None

        pool._cache = {1: job1, 2: job2}
        pool.outbound_buffer.clear()
        pool._active_writers.clear()

        pool.flush()

        job1.discard.assert_called_once()
        job2.discard.assert_not_called()

    @t.skip.if_pypy
    def test_flush_synack_cancels_unaccepted_jobs(self):
        """flush() should call _cancel() on unaccepted jobs when synack is enabled."""
        pool = asynpool.AsynPool(processes=1, synack=True, threads=False)
        pool._state = asynpool.RUN
        pool.maintain_pool = Mock(name='maintain_pool')

        job1 = Mock(name='job1')
        job1._accepted = False
        job1._writer.return_value = None
        job2 = Mock(name='job2')
        job2._accepted = True
        job2._writer.return_value = None

        pool._cache = {1: job1, 2: job2}
        pool.outbound_buffer.clear()
        pool._active_writers.clear()

        pool.flush()

        job1._cancel.assert_called_once()
        job1.discard.assert_not_called()
        job2._cancel.assert_not_called()

    @t.skip.if_pypy
    @patch('billiard.pool.Pool._create_worker_process')
    def test_flush_dead_process_discards_active_writer(self, _create_worker_process):
        """flush() must discard generator from _active_writers when process is dead.

        Previously, when a process was dead, the generator was never removed
        from _active_writers, causing an infinite loop in the while loop.
        """
        pool = asynpool.AsynPool(processes=1, synack=False, threads=False)
        pool._state = asynpool.RUN
        pool.maintain_pool = Mock(name='maintain_pool')

        # Create a mock generator (already started, so not gen_not_started)
        gen = Mock(name='gen')
        gen.__name__ = '_write_job'
        # Simulate a started generator
        with patch.object(asynpool, 'gen_not_started', return_value=False):
            proc = Mock(name='proc')
            proc._is_alive.return_value = False  # Process is dead

            job = Mock(name='job')
            job._accepted = True
            job._write_to = proc
            job._writer.return_value = gen

            pool._cache = {1: job}
            pool._active_writers = {gen}
            pool.outbound_buffer.clear()

            pool.flush()

        # Generator should have been removed from active_writers
        assert gen not in pool._active_writers
        # Job should have been discarded since process is dead
        job.discard.assert_called()

    @t.skip.if_pypy
    @patch('billiard.pool.Pool._create_worker_process')
    def test_flush_alive_process_flushes_writer(self, _create_worker_process):
        """flush() should call _flush_writer when process is still alive."""
        pool = asynpool.AsynPool(processes=1, synack=False, threads=False)
        pool._state = asynpool.RUN
        pool.maintain_pool = Mock(name='maintain_pool')

        gen = Mock(name='gen')
        gen.__name__ = '_write_job'

        with patch.object(asynpool, 'gen_not_started', return_value=False):
            proc = Mock(name='proc')
            proc._is_alive.return_value = True

            job = Mock(name='job')
            job._accepted = True
            job._write_to = proc
            job._writer.return_value = gen

            pool._cache = {1: job}
            pool._active_writers = {gen}
            pool.outbound_buffer.clear()

            with patch.object(pool, '_flush_writer') as mock_flush:
                # _flush_writer removes from _active_writers in its finally
                def side_effect(p, g):
                    pool._active_writers.discard(g)
                mock_flush.side_effect = side_effect

                pool.flush()

            mock_flush.assert_called_once_with(proc, gen)

    @t.skip.if_pypy
    @patch('billiard.pool.Pool._create_worker_process')
    def test_flush_write_ack_coroutine_is_advanced_not_dropped(self, _create_worker_process):
        """flush() must not silently drop _write_ack generators (synack mode).

        _write_ack coroutines are added to _active_writers by send_ack() but
        are NOT mapped in owned_by (which is built from _cache job writers only).
        Dropping them mid-write leaves a partially-written ack on the synq pipe
        and hangs the worker process waiting for the ack that never arrives.
        flush() must advance them to completion instead.
        """
        pool = asynpool.AsynPool(processes=1, synack=True, threads=False)
        pool._state = asynpool.RUN
        pool.maintain_pool = Mock(name='maintain_pool')

        # Simulate a _write_ack generator (name != '_write_job', not in owned_by)
        ack_gen = Mock(name='ack_gen')
        ack_gen.__name__ = '_write_ack'
        # First call to next() returns normally (still writing),
        # second raises StopIteration (write complete).
        ack_gen.__next__ = Mock(side_effect=[None, StopIteration()])

        pool._cache = {}
        pool._active_writers = {ack_gen}
        pool.outbound_buffer.clear()

        pool.flush()

        # Generator should have been advanced (not just silently discarded)
        assert ack_gen.__next__.called
        # And removed once it signalled completion
        assert ack_gen not in pool._active_writers

    @t.skip.if_pypy
    @patch('billiard.pool.Pool._create_worker_process')
    def test_flush_write_ack_coroutine_handles_oserror(self, _create_worker_process):
        """flush() should discard the coroutine if OSError is raised during next()."""
        pool = asynpool.AsynPool(processes=1, synack=True, threads=False)
        pool._state = asynpool.RUN
        pool.maintain_pool = Mock(name='maintain_pool')

        ack_gen = MagicMock(name='ack_gen')
        ack_gen.__name__ = '_write_ack'
        ack_gen.__next__.side_effect = OSError()

        pool._cache = {}
        pool._active_writers = {ack_gen}
        pool.outbound_buffer.clear()
        pool.flush()

        assert ack_gen not in pool._active_writers

    @t.skip.if_pypy
    @patch('billiard.pool.Pool._create_worker_process')
    def test_flush_write_ack_coroutine_handles_eoferror(self, _create_worker_process):
        """flush() should discard the coroutine if EOFError is raised during next()."""
        pool = asynpool.AsynPool(processes=1, synack=True, threads=False)
        pool._state = asynpool.RUN
        pool.maintain_pool = Mock(name='maintain_pool')

        ack_gen = MagicMock(name='ack_gen')
        ack_gen.__name__ = '_write_ack'
        ack_gen.__next__.side_effect = EOFError()

        pool._cache = {}
        pool._active_writers = {ack_gen}
        pool.outbound_buffer.clear()
        pool.flush()

        assert ack_gen not in pool._active_writers

    @t.skip.if_pypy
    @patch('billiard.pool.Pool._create_worker_process')
    def test_flush_not_started_write_job_is_discarded(self, _create_worker_process):
        """flush() should discard a _write_job generator that has not started yet.

        When gen_not_started() returns True the job has not been written to the
        pipe at all, so it is safe to discard it and let the broker redeliver.
        """
        pool = asynpool.AsynPool(processes=1, synack=False, threads=False)
        pool._state = asynpool.RUN
        pool.maintain_pool = Mock(name='maintain_pool')

        gen = Mock(name='gen')
        gen.__name__ = '_write_job'

        with patch.object(asynpool, 'gen_not_started', return_value=True):
            job = Mock(name='job')
            job._accepted = True
            job._writer.return_value = gen

            pool._cache = {1: job}
            pool._active_writers = {gen}
            pool.outbound_buffer.clear()

            pool.flush()

        job.discard.assert_called_once()
        assert gen not in pool._active_writers

    def test_process_result(self):
        x = asynpool.ResultHandler(
            Mock(), Mock(), {}, Mock(),
            Mock(), Mock(), Mock(), Mock(),
            fileno_to_outq={},
            on_process_alive=Mock(),
            on_job_ready=Mock(),
        )
        assert x
        hub = Mock(name='hub')
        recv = x._recv_message = Mock(name='recv_message')
        recv.return_value = iter([])
        x.on_state_change = Mock()
        x.register_with_event_loop(hub)
        proc = x.fileno_to_outq[3] = Mock()
        reader = proc.outq._reader
        reader.poll.return_value = False
        x.handle_event(6)  # KeyError
        x.handle_event(3)
        x._recv_message.assert_called_with(
            hub.add_reader, 3, x.on_state_change,
        )


class test_TaskPool:

    def test_start(self):
        pool = TaskPool(10)
        pool.start()
        assert pool._pool.started
        assert pool._pool._state == asynpool.RUN

        _pool = pool._pool
        pool.stop()
        assert _pool.closed
        assert _pool.joined
        pool.stop()

        pool.start()
        _pool = pool._pool
        pool.terminate()
        pool.terminate()
        assert _pool.terminated

    def test_restart(self):
        pool = TaskPool(10)
        pool._pool = Mock(name='pool')
        pool.restart()
        pool._pool.restart.assert_called_with()
        pool._pool.apply_async.assert_called_with(mp.noop)

    def test_did_start_ok(self):
        pool = TaskPool(10)
        pool._pool = Mock(name='pool')
        assert pool.did_start_ok() is pool._pool.did_start_ok()

    def test_register_with_event_loop(self):
        pool = TaskPool(10)
        pool._pool = Mock(name='pool')
        loop = Mock(name='loop')
        pool.register_with_event_loop(loop)
        pool._pool.register_with_event_loop.assert_called_with(loop)

    def test_on_close(self):
        pool = TaskPool(10)
        pool._pool = Mock(name='pool')
        pool._pool._state = mp.RUN
        pool.on_close()
        pool._pool.close.assert_called_with()

    def test_on_close__pool_not_running(self):
        pool = TaskPool(10)
        pool._pool = Mock(name='pool')
        pool._pool._state = mp.CLOSE
        pool.on_close()
        pool._pool.close.assert_not_called()

    @patch('celery.concurrency.prefork.get_event_loop')
    @patch('celery.concurrency.prefork.threading.Thread')
    def test_on_stop_with_hub_fires_timers(self, mock_thread, mock_get_event_loop):
        pool = TaskPool(10)
        mock_pool = Mock(name='pool')
        mock_pool._state = mp.RUN
        pool._pool = mock_pool

        mock_hub = Mock(name='hub')
        mock_get_event_loop.return_value = mock_hub
        mock_timer_thread = Mock(name='timer_thread')
        mock_thread.return_value = mock_timer_thread

        pool.on_stop()

        mock_pool.close.assert_called_with()
        mock_pool.join.assert_called_with()
        mock_get_event_loop.assert_called_once()
        mock_thread.assert_called_once()
        assert mock_thread.call_args[1]['daemon'] is True
        mock_timer_thread.start.assert_called_once()
        mock_timer_thread.join.assert_called_once_with(timeout=1.0)

    @patch('celery.concurrency.prefork.get_event_loop')
    @patch('celery.concurrency.prefork.threading.Thread')
    @patch('celery.concurrency.prefork.threading.Event')
    def test_on_stop_timer_thread_handles_exceptions(
        self,
        mock_event_class,
        mock_thread,
        mock_get_event_loop,
    ):
        pool = TaskPool(10)
        mock_pool = Mock(name='pool')
        mock_pool._state = mp.RUN
        pool._pool = mock_pool

        mock_hub = Mock(name='hub')
        mock_hub.fire_timers.side_effect = [Exception("Hub error"), None]
        mock_get_event_loop.return_value = mock_hub

        mock_shutdown_event = Mock(name='shutdown_event')
        # Simulate two loop iterations and then shutdown
        mock_shutdown_event.is_set.side_effect = [False, False, True]
        mock_event_class.return_value = mock_shutdown_event

        thread_target = None

        def capture_thread(*args, **kwargs):
            nonlocal thread_target
            thread_target = kwargs['target']
            mock_timer_thread = Mock(name='timer_thread')
            return mock_timer_thread

        mock_thread.side_effect = capture_thread

        pool.on_stop()

        with patch('celery.concurrency.prefork.time.sleep'):
            thread_target()

        # Should match number of loop iterations allowed by mock_shutdown_event.is_set.side_effect
        assert mock_hub.fire_timers.call_count == 2

    @patch('celery.concurrency.prefork.get_event_loop')
    def test_on_stop_no_hub(self, mock_get_event_loop):
        pool = TaskPool(10)
        mock_pool = Mock(name='pool')
        mock_pool._state = mp.RUN
        pool._pool = mock_pool

        mock_get_event_loop.return_value = None

        pool.on_stop()

        mock_pool.close.assert_called_with()
        mock_pool.join.assert_called_with()
        mock_get_event_loop.assert_called_once()

    def test_apply_async(self):
        pool = TaskPool(10)
        pool.start()
        pool.apply_async(lambda x: x, (2,), {})

    def test_grow_shrink(self):
        pool = TaskPool(10)
        pool.start()
        assert pool._pool._processes == 10
        pool.grow()
        assert pool._pool._processes == 11
        pool.shrink(2)
        assert pool._pool._processes == 9

    def test_info(self):
        pool = TaskPool(10)
        procs = [Bunch(pid=i) for i in range(pool.limit)]

        class _Pool:
            _pool = procs
            _maxtasksperchild = None
            timeout = 10
            soft_timeout = 5

            def human_write_stats(self, *args, **kwargs):
                return {}
        pool._pool = _Pool()
        info = pool.info
        assert info['max-concurrency'] == pool.limit
        assert info['max-tasks-per-child'] == 'N/A'
        assert info['timeouts'] == (5, 10)

    def test_num_processes(self):
        pool = TaskPool(7)
        pool.start()
        assert pool.num_processes == 7

    @patch('billiard.forking_enable')
    def test_on_start_proc_alive_timeout_default(self, __forking_enable):
        app = Mock(conf=AttributeDict(DEFAULTS))
        pool = TaskPool(4, app=app)
        pool.on_start()
        assert pool._pool._proc_alive_timeout == 4.0

    @patch('billiard.forking_enable')
    def test_on_start_proc_alive_timeout_custom(self, __forking_enable):
        app = Mock(conf=AttributeDict(DEFAULTS))
        app.conf.worker_proc_alive_timeout = 8.0
        pool = TaskPool(4, app=app)
        pool.on_start()
        assert pool._pool._proc_alive_timeout == 8.0



"""
Integration tests for RabbitMQ cycle detection bug with quorum queues.

This reproduces the conditions for
https://github.com/celery/celery/issues/9867 when running against RabbitMQ
<4.0.1.
"""

import os

import pytest
from kombu import Exchange, Queue

from celery import Celery, exceptions
from celery.contrib.testing.worker import start_worker


class TaskFailedException(Exception):
    """Test exception for tasks that should fail and retry."""
    pass


def get_rabbitmq_url():
    """Get RabbitMQ broker URL from environment."""
    user = os.environ.get("RABBITMQ_DEFAULT_USER", "guest")
    password = os.environ.get("RABBITMQ_DEFAULT_PASSWORD", "guest")
    return os.environ.get(
        "TEST_BROKER", f"pyamqp://{user}:{password}@localhost:5672//")


def get_backend_url():
    """Get backend URL from environment."""
    return os.environ.get("TEST_BACKEND", "rpc")


def create_test_app_with_quorum_queues():
    """
    Create Celery app configured for quorum queue testing.

    Returns:
        Tuple of (app, queue_name)
    """
    broker_url = get_rabbitmq_url()
    backend_url = get_backend_url()

    app = Celery(
        "test_cycle_detection",
        broker=broker_url,
        backend=backend_url,
    )

    # Configure queue with quorum queue type
    queue_name = 'test_cycle_queue'
    default_exchange = Exchange('celery.topic', type='topic')

    app.conf.task_queues = [
        Queue(
            queue_name,
            exchange=default_exchange,
            routing_key=queue_name,
            queue_arguments={'x-queue-type': 'quorum'}
        ),
    ]

    app.conf.task_default_queue = queue_name

    # Recommended settings for quorum queues
    app.conf.broker_transport_options = {"confirm_publish": True}

    # Enable result extended to track retries
    app.conf.result_extended = True

    return app


@pytest.mark.amqp
@pytest.mark.timeout(35)
def test_countdown_task_with_retry():
    """Test where a countdown task gets retried with the same delay."""
    app = create_test_app_with_quorum_queues()

    @app.task(bind=True, default_retry_delay=3, max_retries=1)
    def countdown_task_with_retry(self):
        """Task that fails once then succeeds."""
        # First attempt: fail to trigger retry
        if self.request.retries < 1:
            raise self.retry(exc=TaskFailedException("Simulated failure"))

        return self.request.retries

    with start_worker(
        app,
        loglevel="INFO",
        shutdown_timeout=15,
    ):
        # Execute task with countdown=3 (same as default_retry_delay)
        # We need to hit the same delayed buckets to trigger the cycle detection
        # in RabbitMQ.
        result = countdown_task_with_retry.apply_async(countdown=3)

        try:
            # Wait for the task to complete, we must wait at least long enough
            # for the countdown to be reached and the retry to occur (6s).
            value = result.get(timeout=15)
        except exceptions.TimeoutError as e:
            pytest.fail(
                f"Task was silently dropped by RabbitMQ due to cycle detection. "
                f"Exception: {e!r}."
            )

        assert value == 1, \
            f"Expected task to succeed after 1 retry, got: {value}"


@pytest.mark.amqp
@pytest.mark.timeout(35)
def test_non_eta_task_with_multiple_retries():
    """Test where a task gets retried twice with the same delay."""
    app = create_test_app_with_quorum_queues()

    @app.task(bind=True, max_retries=2)
    def non_eta_task_with_retries(self):
        """Task that fails twice then succeeds."""
        # First and second attempts: fail to trigger retry
        if self.request.retries < 2:
            raise self.retry(exc=TaskFailedException("Simulated failure"), countdown=3)

        return self.request.retries

    with start_worker(
        app,
        loglevel="INFO",
        shutdown_timeout=15,
    ):
        result = non_eta_task_with_retries.apply_async()

        try:
            # Wait for the task to complete, we must wait at least long enough
            # for the two retries to occur (6s).
            value = result.get(timeout=15)
        except exceptions.TimeoutError as e:
            pytest.fail(
                f"Task was silently dropped by RabbitMQ due to cycle detection. "
                f"Exception: {e!r}"
            )

        assert value == 2, \
            f"Expected task to succeed after 2 retries, got: {value}"



import os
import re
from datetime import datetime, timedelta, timezone
from time import sleep
from unittest.mock import ANY

import pytest

from celery.utils.nodenames import anon_nodename

from .tasks import add, sleeping

NODENAME = anon_nodename()

_flaky = pytest.mark.flaky(reruns=5, reruns_delay=2)
_timeout = pytest.mark.timeout(timeout=300)


def flaky(fn):
    return _timeout(_flaky(fn))


@pytest.fixture()
def inspect(manager):
    return manager.app.control.inspect()


class test_Inspect:
    """Integration tests to app.control.inspect() API"""

    @flaky
    def test_ping(self, inspect):
        """Tests pinging the worker"""
        ret = inspect.ping()
        assert len(ret) == 1
        assert ret[NODENAME] == {'ok': 'pong'}
        # TODO: Check ping() is returning None after stopping worker.
        # This is tricky since current test suite does not support stopping of
        # the worker.

    @flaky
    def test_clock(self, inspect):
        """Tests getting clock information from worker"""
        ret = inspect.clock()
        assert len(ret) == 1
        assert ret[NODENAME]['clock'] > 0

    @flaky
    def test_registered(self, inspect):
        """Tests listing registered tasks"""
        # TODO: We can check also the exact values of the registered methods
        ret = inspect.registered()
        assert len(ret) == 1
        assert len(ret[NODENAME]) > 0
        for task_name in ret[NODENAME]:
            assert isinstance(task_name, str)

        ret = inspect.registered('name')
        for task_info in ret[NODENAME]:
            # task_info is in form 'TASK_NAME [name=TASK_NAME]'
            assert re.fullmatch(r'\S+ \[name=\S+\]', task_info)

    @flaky
    def test_active_queues(self, inspect):
        """Tests listing active queues"""
        ret = inspect.active_queues()
        assert len(ret) == 1
        assert ret[NODENAME] == [
            {
                'alias': None,
                'auto_delete': False,
                'binding_arguments': None,
                'bindings': [],
                'consumer_arguments': None,
                'durable': True,
                'exchange': {
                    'arguments': None,
                    'auto_delete': False,
                    'delivery_mode': None,
                    'durable': True,
                    'name': 'celery',
                    'no_declare': False,
                    'passive': False,
                    'type': 'direct'
                },
                'exclusive': False,
                'expires': None,
                'max_length': None,
                'max_length_bytes': None,
                'max_priority': None,
                'message_ttl': None,
                'name': 'celery',
                'no_ack': False,
                'no_declare': None,
                'queue_arguments': None,
                'routing_key': 'celery'}
        ]

    @flaky
    def test_active(self, inspect):
        """Tests listing active tasks"""
        res = sleeping.delay(5)
        sleep(1)
        ret = inspect.active()
        assert len(ret) == 1
        assert ret[NODENAME] == [
            {
                'id': res.task_id,
                'name': 't.integration.tasks.sleeping',
                'args': [5],
                'kwargs': {},
                'type': 't.integration.tasks.sleeping',
                'hostname': ANY,
                'time_start': ANY,
                'acknowledged': True,
                'delivery_info': {
                    'exchange': '',
                    'routing_key': 'celery',
                    'priority': 0,
                    'redelivered': False
                },
                'worker_pid': ANY
            }
        ]

    @flaky
    def test_scheduled(self, inspect):
        """Tests listing scheduled tasks"""
        exec_time = datetime.now(timezone.utc) + timedelta(seconds=5)
        res = add.apply_async([1, 2], {'z': 3}, eta=exec_time)
        ret = inspect.scheduled()
        assert len(ret) == 1
        assert ret[NODENAME] == [
            {
                'eta': exec_time.strftime('%Y-%m-%dT%H:%M:%S.%f') + '+00:00',
                'priority': 6,
                'request': {
                    'id': res.task_id,
                    'name': 't.integration.tasks.add',
                    'args': [1, 2],
                    'kwargs': {'z': 3},
                    'type': 't.integration.tasks.add',
                    'hostname': ANY,
                    'time_start': None,
                    'acknowledged': False,
                    'delivery_info': {
                        'exchange': '',
                        'routing_key': 'celery',
                        'priority': 0,
                        'redelivered': False
                    },
                    'worker_pid': None
                }
            }
        ]

    @flaky
    def test_query_task(self, inspect):
        """Task that does not exist or is finished"""
        ret = inspect.query_task('d08b257e-a7f1-4b92-9fea-be911441cb2a')
        assert len(ret) == 1
        assert ret[NODENAME] == {}

        # Task in progress
        res = sleeping.delay(5)
        sleep(1)
        ret = inspect.query_task(res.task_id)
        assert len(ret) == 1
        assert ret[NODENAME] == {
            res.task_id: [
                'active', {
                    'id': res.task_id,
                    'name': 't.integration.tasks.sleeping',
                    'args': [5],
                    'kwargs': {},
                    'type': 't.integration.tasks.sleeping',
                    'hostname': NODENAME,
                    'time_start': ANY,
                    'acknowledged': True,
                    'delivery_info': {
                        'exchange': '',
                        'routing_key': 'celery',
                        'priority': 0,
                        'redelivered': False
                    },
                    # worker is running in the same process as separate thread
                    'worker_pid': ANY
                }
            ]
        }

    @flaky
    def test_stats(self, inspect):
        """tests fetching statistics"""
        ret = inspect.stats()
        assert len(ret) == 1
        assert ret[NODENAME]['pool']['max-concurrency'] == 1
        assert len(ret[NODENAME]['pool']['processes']) == 1
        assert ret[NODENAME]['uptime'] > 0
        # worker is running in the same process as separate thread
        assert ret[NODENAME]['pid'] == os.getpid()

    @flaky
    def test_report(self, inspect):
        """Tests fetching report"""
        ret = inspect.report()
        assert len(ret) == 1
        assert ret[NODENAME] == {'ok': ANY}

    @flaky
    def test_revoked(self, inspect):
        """Testing revoking of task"""
        # Fill the queue with tasks to fill the queue
        for _ in range(4):
            sleeping.delay(2)
        # Execute task and revoke it
        result = add.apply_async((1, 1))
        result.revoke()
        ret = inspect.revoked()
        assert len(ret) == 1
        assert result.task_id in ret[NODENAME]

    @flaky
    def test_conf(self, inspect):
        """Tests getting configuration"""
        ret = inspect.conf()
        assert len(ret) == 1
        assert ret[NODENAME]['worker_hijack_root_logger'] == ANY
        assert ret[NODENAME]['worker_log_color'] == ANY
        assert ret[NODENAME]['accept_content'] == ANY
        assert ret[NODENAME]['enable_utc'] == ANY
        assert ret[NODENAME]['timezone'] == ANY
        assert ret[NODENAME]['broker_url'] == ANY
        assert ret[NODENAME]['result_backend'] == ANY
        assert ret[NODENAME]['broker_heartbeat'] == ANY
        assert ret[NODENAME]['deprecated_settings'] == ANY
        assert ret[NODENAME]['include'] == ANY



import subprocess

import pytest

from celery import Celery


def test_run_worker():
    with pytest.raises(subprocess.CalledProcessError) as exc_info:
        subprocess.check_output(
            ["celery", "--config", "t.integration.test_worker_config", "worker"],
            stderr=subprocess.STDOUT)

    called_process_error = exc_info.value
    assert called_process_error.returncode == 1, called_process_error
    output = called_process_error.output.decode('utf-8')
    assert output.find(
        "Retrying to establish a connection to the message broker after a connection "
        "loss has been disabled (app.conf.broker_connection_retry_on_startup=False). "
        "Shutting down...") != -1, output


def test_django_fixup_direct_worker(caplog, monkeypatch):
    """Test Django fixup by directly instantiating Celery worker without subprocess."""
    import logging

    import django

    # Set logging level to capture debug messages
    caplog.set_level(logging.DEBUG)

    # Configure Django settings
    monkeypatch.setenv('DJANGO_SETTINGS_MODULE', 't.integration.test_django_settings')
    django.setup()

    # Create Celery app with Django integration
    app = Celery('test_django_direct')
    app.config_from_object('django.conf:settings', namespace='CELERY')
    app.autodiscover_tasks()

    # Test that we can access worker configuration without recursion errors
    # This should trigger the Django fixup initialization
    worker = app.Worker(
        pool='solo',
        concurrency=1,
        loglevel='debug'
    )

    # Accessing pool_cls should not cause AttributeError
    pool_cls = worker.pool_cls
    assert pool_cls is not None

    # Verify pool_cls has __module__ attribute (should be a class, not a string)
    assert hasattr(pool_cls, '__module__'), \
        f"pool_cls should be a class with __module__, got {type(pool_cls)}: {pool_cls}"

    # Capture and check logs
    log_output = caplog.text

    # Verify no recursion-related errors in logs
    assert "RecursionError" not in log_output, f"RecursionError found in logs:\n{log_output}"
    assert "maximum recursion depth exceeded" not in log_output, \
        f"Recursion depth error found in logs:\n{log_output}"

    assert "AttributeError: 'str' object has no attribute '__module__'." not in log_output, \
        f"AttributeError found in logs:\n{log_output}"



import threading
from contextlib import contextmanager
from unittest.mock import patch

import pytest

import celery.contrib.testing.worker as contrib_embed_worker
from celery.app import backends
from celery.backends.cache import CacheBackend
from celery.exceptions import ImproperlyConfigured
from celery.utils.nodenames import anon_nodename


class CachedBackendWithTreadTrucking(CacheBackend):
    test_instance_count = 0
    test_call_stats = {}

    def _track_attribute_access(self, method_name):
        cls = type(self)

        instance_no = getattr(self, '_instance_no', None)
        if instance_no is None:
            instance_no = self._instance_no = cls.test_instance_count
            cls.test_instance_count += 1
            cls.test_call_stats[instance_no] = []

        cls.test_call_stats[instance_no].append({
            'thread_id': threading.get_ident(),
            'method_name': method_name
        })

    def __getattribute__(self, name):
        if name == '_instance_no' or name == '_track_attribute_access':
            return super().__getattribute__(name)

        if name.startswith('__') and name != '__init__':
            return super().__getattribute__(name)

        self._track_attribute_access(name)
        return super().__getattribute__(name)


@contextmanager
def embed_worker(app,
                 concurrency=1,
                 pool='threading', **kwargs):
    """
    Helper embedded worker for testing.

    It's based on a :func:`celery.contrib.testing.worker.start_worker`,
    but doesn't modify logging settings and additionally shutdown
    worker pool.
    """
    # prepare application for worker
    app.finalize()
    app.set_current()

    worker = contrib_embed_worker.TestWorkController(
        app=app,
        concurrency=concurrency,
        hostname=anon_nodename(),
        pool=pool,
        # not allowed to override TestWorkController.on_consumer_ready
        ready_callback=None,
        without_heartbeat=kwargs.pop("without_heartbeat", True),
        without_mingle=True,
        without_gossip=True,
        **kwargs
    )

    t = threading.Thread(target=worker.start, daemon=True)
    t.start()
    worker.ensure_started()

    yield worker

    worker.stop()
    t.join(10.0)
    if t.is_alive():
        raise RuntimeError(
            "Worker thread failed to exit within the allocated timeout. "
            "Consider raising `shutdown_timeout` if your tasks take longer "
            "to execute."
        )


class test_backends:

    @pytest.mark.parametrize('url,expect_cls', [
        ('cache+memory://', CacheBackend),
    ])
    def test_get_backend_aliases(self, url, expect_cls, app):
        backend, url = backends.by_url(url, app.loader)
        assert isinstance(backend(app=app, url=url), expect_cls)

    def test_unknown_backend(self, app):
        with pytest.raises(ImportError):
            backends.by_name('fasodaopjeqijwqe', app.loader)

    def test_backend_by_url(self, app, url='redis://localhost/1'):
        from celery.backends.redis import RedisBackend
        backend, url_ = backends.by_url(url, app.loader)
        assert backend is RedisBackend
        assert url_ == url

    def test_sym_raises_ValuError(self, app):
        with patch('celery.app.backends.symbol_by_name') as sbn:
            sbn.side_effect = ValueError()
            with pytest.raises(ImproperlyConfigured):
                backends.by_name('xxx.xxx:foo', app.loader)

    def test_backend_can_not_be_module(self, app):
        with pytest.raises(ImproperlyConfigured):
            backends.by_name(pytest, app.loader)

    @pytest.mark.celery(
        result_backend=f'{CachedBackendWithTreadTrucking.__module__}.'
        f'{CachedBackendWithTreadTrucking.__qualname__}'
        f'+memory://')
    def test_backend_thread_safety(self):
        @self.app.task
        def dummy_add_task(x, y):
            return x + y

        with embed_worker(app=self.app, pool='threads'):
            result = dummy_add_task.delay(6, 9)
            assert result.get(timeout=10) == 15

        call_stats = CachedBackendWithTreadTrucking.test_call_stats
        # check that backend instance is used without same thread
        for backend_call_stats in call_stats.values():
            thread_ids = set()
            for call_stat in backend_call_stats:
                thread_ids.add(call_stat['thread_id'])
            assert len(thread_ids) <= 1, \
                "The same celery backend instance is used by multiple threads"



from __future__ import annotations

from unittest.mock import patch

import pytest
from click.testing import CliRunner

from celery.bin.celery import celery

from .proj import daemon_config as config


@pytest.mark.usefixtures('depends_on_current_app')
@pytest.mark.parametrize("daemon", ["worker", "beat", "events"])
def test_daemon_options_from_config(daemon: str, cli_runner: CliRunner):

    with patch(f"celery.bin.{daemon}.{daemon}.callback") as mock:
        cli_runner.invoke(celery, f"-A t.unit.bin.proj.daemon {daemon}")

    mock.assert_called_once()
    for param in "logfile", "pidfile", "uid", "gid", "umask", "executable":
        assert mock.call_args.kwargs[param] == getattr(config, f"{daemon}_{param}")



import logging
import platform
import time
from datetime import datetime, timedelta, timezone
from uuid import uuid4

import billiard as multiprocessing
import pytest

import celery
from celery import chain, chord, group
from celery.canvas import StampingVisitor
from celery.signals import task_received
from celery.utils.serialization import UnpickleableExceptionWrapper
from celery.worker import state as worker_state

from .conftest import TEST_BACKEND, get_active_redis_channels, get_redis_connection
from .tasks import (ClassBasedAutoRetryTask, ExpectedException, add, add_ignore_result, add_not_typed, add_pydantic,
                    add_pydantic_string_annotations, fail, fail_unpickleable, print_unicode, retry, retry_once,
                    retry_once_headers, retry_once_priority, retry_unpickleable, return_properties,
                    second_order_replace1, sleeping, soft_time_limit_must_exceed_time_limit)

TIMEOUT = 10

_flaky = pytest.mark.flaky(reruns=5, reruns_delay=2)
_timeout = pytest.mark.timeout(timeout=300)


def flaky(fn):
    return _timeout(_flaky(fn))


def set_multiprocessing_start_method():
    """Set multiprocessing start method to 'fork' if not on Linux."""
    if platform.system() != "Linux":
        try:
            multiprocessing.set_start_method("fork")
        except RuntimeError:
            # The method is already set
            pass


class test_class_based_tasks:

    @flaky
    def test_class_based_task_retried(self, celery_session_app,
                                      celery_session_worker):
        task = ClassBasedAutoRetryTask()
        celery_session_app.register_task(task)
        res = task.delay()
        assert res.get(timeout=TIMEOUT) == 1


def _producer(j):
    """Single producer helper function"""
    results = []
    for i in range(20):
        results.append([i + j, add.delay(i, j)])
    for expected, result in results:
        value = result.get(timeout=10)
        assert value == expected
        assert result.status == 'SUCCESS'
        assert result.ready() is True
        assert result.successful() is True
    return j


class test_tasks:

    def test_simple_call(self):
        """Tests direct simple call of task"""
        assert add(1, 1) == 2
        assert add(1, 1, z=1) == 3

    @flaky
    def test_basic_task(self, manager):
        """Tests basic task call"""
        results = []
        # Tests calling task only with args
        for i in range(10):
            results.append([i + i, add.delay(i, i)])
        for expected, result in results:
            value = result.get(timeout=10)
            assert value == expected
            assert result.status == 'SUCCESS'
            assert result.ready() is True
            assert result.successful() is True

        results = []
        # Tests calling task with args and kwargs
        for i in range(10):
            results.append([3*i, add.delay(i, i, z=i)])
        for expected, result in results:
            value = result.get(timeout=10)
            assert value == expected
            assert result.status == 'SUCCESS'
            assert result.ready() is True
            assert result.successful() is True

    @flaky
    @pytest.mark.skip(reason="Broken test")
    def test_multiprocess_producer(self, manager):
        """Testing multiple processes calling tasks."""
        set_multiprocessing_start_method()

        from multiprocessing import Pool
        pool = Pool(20)
        ret = pool.map(_producer, range(120))
        assert list(ret) == list(range(120))

    @flaky
    @pytest.mark.skip(reason="Broken test")
    def test_multithread_producer(self, manager):
        """Testing multiple threads calling tasks."""
        set_multiprocessing_start_method()

        from multiprocessing.pool import ThreadPool
        pool = ThreadPool(20)
        ret = pool.map(_producer, range(120))
        assert list(ret) == list(range(120))

    @flaky
    def test_ignore_result(self, manager):
        """Testing calling task with ignoring results."""
        result = add.apply_async((1, 2), ignore_result=True)
        assert result.get() is None
        # We wait since it takes a bit of time for the result to be
        # persisted in the result backend.
        time.sleep(1)
        assert result.result is None

    @flaky
    def test_pydantic_annotations(self, manager):
        """Tests task call with Pydantic model serialization."""
        results = []
        # Tests calling task only with args
        for i in range(10):
            results.append([i + i, add_pydantic.delay({'x': i, 'y': i})])
        for expected, result in results:
            value = result.get(timeout=10)
            assert value == {'result': expected}
            assert result.status == 'SUCCESS'
            assert result.ready() is True
            assert result.successful() is True

    @flaky
    def test_pydantic_string_annotations(self, manager):
        """Tests task call with string-annotated Pydantic model."""
        results = []
        # Tests calling task only with args
        for i in range(10):
            results.append([i + i, add_pydantic_string_annotations.delay({'x': i, 'y': i})])
        for expected, result in results:
            value = result.get(timeout=10)
            assert value == {'result': expected}
            assert result.status == 'SUCCESS'
            assert result.ready() is True
            assert result.successful() is True

    @flaky
    def test_timeout(self, manager):
        """Testing timeout of getting results from tasks."""
        result = sleeping.delay(10)
        with pytest.raises(celery.exceptions.TimeoutError):
            result.get(timeout=5)

    @pytest.mark.timeout(60)
    @flaky
    def test_expired(self, manager):
        """Testing expiration of task."""
        # Fill the queue with tasks which took > 1 sec to process
        for _ in range(4):
            sleeping.delay(2)
        # Execute task with expiration = 1 sec
        result = add.apply_async((1, 1), expires=1)
        with pytest.raises(celery.exceptions.TaskRevokedError):
            result.get()
        assert result.status == 'REVOKED'
        assert result.ready() is True
        assert result.failed() is False
        assert result.successful() is False

        # Fill the queue with tasks which took > 1 sec to process
        for _ in range(4):
            sleeping.delay(2)
        # Execute task with expiration at now + 1 sec
        result = add.apply_async((1, 1), expires=datetime.now(timezone.utc) + timedelta(seconds=1))
        with pytest.raises(celery.exceptions.TaskRevokedError):
            result.get()
        assert result.status == 'REVOKED'
        assert result.ready() is True
        assert result.failed() is False
        assert result.successful() is False

    @flaky
    def test_eta(self, manager):
        """Tests tasks scheduled at some point in future."""
        start = time.perf_counter()
        # Schedule task to be executed in 3 seconds
        result = add.apply_async((1, 1), countdown=3)
        time.sleep(1)
        assert result.status == 'PENDING'
        assert result.ready() is False
        assert result.get() == 2
        end = time.perf_counter()
        assert result.status == 'SUCCESS'
        assert result.ready() is True
        # Difference between calling the task and result must be bigger than 3 secs
        assert (end - start) > 3

        start = time.perf_counter()
        # Schedule task to be executed at time now + 3 seconds
        result = add.apply_async((2, 2), eta=datetime.now(timezone.utc) + timedelta(seconds=3))
        time.sleep(1)
        assert result.status == 'PENDING'
        assert result.ready() is False
        assert result.get() == 4
        end = time.perf_counter()
        assert result.status == 'SUCCESS'
        assert result.ready() is True
        # Difference between calling the task and result must be bigger than 3 secs
        assert (end - start) > 3

    @flaky
    def test_fail(self, manager):
        """Tests that the failing task propagates back correct exception."""
        result = fail.delay()
        with pytest.raises(ExpectedException):
            result.get(timeout=5)
        assert result.status == 'FAILURE'
        assert result.ready() is True
        assert result.failed() is True
        assert result.successful() is False

    @flaky
    def test_revoked(self, manager):
        """Testing revoking of task"""
        # Fill the queue with tasks to fill the queue
        for _ in range(4):
            sleeping.delay(2)
        # Execute task and revoke it
        result = add.apply_async((1, 1))
        result.revoke()
        with pytest.raises(celery.exceptions.TaskRevokedError):
            result.get()
        assert result.status == 'REVOKED'
        assert result.ready() is True
        assert result.failed() is False
        assert result.successful() is False

    def test_revoked_by_headers_simple_canvas(self, manager):
        """Testing revoking of task using a stamped header"""
        target_monitoring_id = uuid4().hex

        class MonitoringIdStampingVisitor(StampingVisitor):
            def on_signature(self, sig, **headers) -> dict:
                return {'monitoring_id': target_monitoring_id}

        for monitoring_id in [target_monitoring_id, uuid4().hex, 4242, None]:
            stamped_task = add.si(1, 1)
            stamped_task.stamp(visitor=MonitoringIdStampingVisitor())
            result = stamped_task.freeze()
            result.revoke_by_stamped_headers(headers={'monitoring_id': [monitoring_id]})
            stamped_task.apply_async()
            if monitoring_id == target_monitoring_id:
                with pytest.raises(celery.exceptions.TaskRevokedError):
                    result.get()
                assert result.status == 'REVOKED'
                assert result.ready() is True
                assert result.failed() is False
                assert result.successful() is False
            else:
                assert result.get() == 2
                assert result.status == 'SUCCESS'
                assert result.ready() is True
                assert result.failed() is False
                assert result.successful() is True

            # Clear the set of revoked stamps in the worker state.
            # This step is performed in each iteration of the loop to ensure that only tasks
            # stamped with a specific monitoring ID will be revoked.
            # For subsequent iterations with different monitoring IDs, the revoked stamps will
            # not match the task's stamps, allowing those tasks to proceed successfully.
            worker_state.revoked_stamps.clear()

    @pytest.mark.timeout(20)
    @pytest.mark.flaky(reruns=2)
    def test_revoked_by_headers_complex_canvas(self, manager, subtests):
        """Testing revoking of task using a stamped header"""
        try:
            manager.app.backend.ensure_chords_allowed()
        except NotImplementedError as e:
            raise pytest.skip(e.args[0])

        for monitoring_id in ["4242", [1234, uuid4().hex]]:

            # Try to purge the queue before we start
            # to attempt to avoid interference from other tests
            manager.wait_until_idle()

            target_monitoring_id = isinstance(monitoring_id, list) and monitoring_id[0] or monitoring_id

            class MonitoringIdStampingVisitor(StampingVisitor):
                def on_signature(self, sig, **headers) -> dict:
                    return {'monitoring_id': target_monitoring_id, 'stamped_headers': ['monitoring_id']}

            stamped_task = sleeping.si(4)
            stamped_task.stamp(visitor=MonitoringIdStampingVisitor())
            result = stamped_task.freeze()

            canvas = [
                group([stamped_task]),
                chord(group([stamped_task]), sleeping.si(2)),
                chord(group([sleeping.si(2)]), stamped_task),
                chain(stamped_task),
                group([sleeping.si(2), stamped_task, sleeping.si(2)]),
                chord([sleeping.si(2), stamped_task], sleeping.si(2)),
                chord([sleeping.si(2), sleeping.si(2)], stamped_task),
                chain(sleeping.si(2), stamped_task),
                chain(sleeping.si(2), group([sleeping.si(2), stamped_task, sleeping.si(2)])),
                chain(sleeping.si(2), group([sleeping.si(2), stamped_task]), sleeping.si(2)),
                chain(sleeping.si(2), group([sleeping.si(2), sleeping.si(2)]), stamped_task),
            ]

            result.revoke_by_stamped_headers(headers={'monitoring_id': monitoring_id})

            for sig in canvas:
                sig_result = sig.apply_async()
                with subtests.test(msg='Testing if task was revoked'):
                    with pytest.raises(celery.exceptions.TaskRevokedError):
                        sig_result.get()
                    assert result.status == 'REVOKED'
                    assert result.ready() is True
                    assert result.failed() is False
                    assert result.successful() is False
            worker_state.revoked_stamps.clear()

    @flaky
    def test_revoke_by_stamped_headers_no_match(self, manager):
        response = manager.app.control.revoke_by_stamped_headers(
            {"myheader": ["myvalue"]},
            terminate=False,
            reply=True,
        )

        expected_response = "headers {'myheader': ['myvalue']} flagged as revoked, but not terminated"
        assert response[0][list(response[0].keys())[0]]["ok"] == expected_response

    @flaky
    def test_wrong_arguments(self, manager):
        """Tests that proper exceptions are raised when task is called with wrong arguments."""
        with pytest.raises(TypeError):
            add(5)

        with pytest.raises(TypeError):
            add(5, 5, wrong_arg=5)

        with pytest.raises(TypeError):
            add.delay(5)

        with pytest.raises(TypeError):
            add.delay(5, wrong_arg=5)

        # Tasks with typing=False are not checked but execution should fail
        result = add_not_typed.delay(5)
        with pytest.raises(TypeError):
            result.get(timeout=5)
        assert result.status == 'FAILURE'

        result = add_not_typed.delay(5, wrong_arg=5)
        with pytest.raises(TypeError):
            result.get(timeout=5)
        assert result.status == 'FAILURE'

    @pytest.mark.xfail(
        condition=TEST_BACKEND == "rpc",
        reason="Retry failed on rpc backend",
        strict=False,
    )
    def test_retry(self, manager):
        """Tests retrying of task."""
        # Tests when max. retries is reached
        result = retry.delay()

        tik = time.monotonic()
        while time.monotonic() < tik + 5:
            status = result.status
            if status != 'PENDING':
                break
            time.sleep(0.1)
        else:
            raise AssertionError("Timeout while waiting for the task to be retried")
        assert status == 'RETRY'
        with pytest.raises(ExpectedException):
            result.get()
        assert result.status == 'FAILURE'

        # Tests when task is retried but after returns correct result
        result = retry.delay(return_value='bar')

        tik = time.monotonic()
        while time.monotonic() < tik + 5:
            status = result.status
            if status != 'PENDING':
                break
            time.sleep(0.1)
        else:
            raise AssertionError("Timeout while waiting for the task to be retried")
        assert status == 'RETRY'
        assert result.get() == 'bar'
        assert result.status == 'SUCCESS'

    def test_retry_with_unpickleable_exception(self, manager):
        """Test a task that retries with an unpickleable exception.

        We expect to be able to fetch the result (exception) correctly.
        """

        job = retry_unpickleable.delay(
            "foo",
            "bar",
            retry_kwargs={"countdown": 10, "max_retries": 1},
        )

        # Wait for the task to raise the Retry exception
        tik = time.monotonic()
        while time.monotonic() < tik + 5:
            status = job.status
            if status != 'PENDING':
                break
            time.sleep(0.1)
        else:
            raise AssertionError("Timeout while waiting for the task to be retried")

        assert status == 'RETRY'

        # Get the exception
        res = job.result
        assert job.status == 'RETRY'  # make sure that it wasn't completed yet

        # Check it. Accept both the dedicated wrapper and plain Exception
        # (some environments may return a stringified Exception instead).
        if isinstance(res, UnpickleableExceptionWrapper):
            assert res.exc_cls_name == "UnpickleableException"
            assert res.exc_args == ("foo",)
        else:
            # Fallback: ensure the exception string mentions the class and argument
            res_str = str(res)
            assert "UnpickleableException" in res_str
            assert "foo" in res_str

        job.revoke()

    def test_fail_with_unpickleable_exception(self, manager):
        """Test a task that fails with an unpickleable exception.

        We expect to be able to fetch the result (exception) correctly.
        """
        result = fail_unpickleable.delay("foo", "bar")

        # Accept either the dedicated wrapper exception or a plain Exception
        # whose string contains the class name and args (some backends
        # may return a stringified exception).
        try:
            result.get()
            pytest.fail("Expected an exception when getting result")
        except UnpickleableExceptionWrapper as exc_wrapper:
            assert exc_wrapper.exc_cls_name == "UnpickleableException"
            assert exc_wrapper.exc_args == ("foo",)
        except Exception as exc:
            # Fallback: ensure the exception string mentions the class and argument
            exc_str = str(exc)
            assert "UnpickleableException" in exc_str
            assert "foo" in exc_str

        assert result.status == 'FAILURE'

    # Requires investigation why it randomly succeeds/fails
    @pytest.mark.skip(reason="Randomly fails")
    def test_task_accepted(self, manager, sleep=1):
        r1 = sleeping.delay(sleep)
        sleeping.delay(sleep)
        manager.assert_accepted([r1.id])

    @flaky
    def test_task_retried_once(self, manager):
        res = retry_once.delay()
        assert res.get(timeout=TIMEOUT) == 1  # retried once

    @flaky
    def test_task_retried_once_with_expires(self, manager):
        res = retry_once.delay(expires=60)
        assert res.get(timeout=TIMEOUT) == 1  # retried once

    @flaky
    def test_task_retried_priority(self, manager):
        res = retry_once_priority.apply_async(priority=7)
        assert res.get(timeout=TIMEOUT) == 7  # retried once with priority 7

    @flaky
    def test_task_retried_headers(self, manager):
        res = retry_once_headers.apply_async(headers={'x-test-header': 'test-value'})
        headers = res.get(timeout=TIMEOUT)
        assert headers is not None  # retried once with headers
        assert 'x-test-header' in headers  # retry keeps custom headers

    @flaky
    def test_unicode_task(self, manager):
        manager.join(
            group(print_unicode.s() for _ in range(5))(),
            timeout=TIMEOUT, propagate=True,
        )

    @flaky
    def test_properties(self, celery_session_worker):
        res = return_properties.apply_async(app_id="1234")
        assert res.get(timeout=TIMEOUT)["app_id"] == "1234"

    @flaky
    def test_soft_time_limit_exceeding_time_limit(self):

        with pytest.raises(ValueError, match='soft_time_limit must be less than or equal to time_limit'):
            result = soft_time_limit_must_exceed_time_limit.apply_async()
            result.get(timeout=5)

            assert result.status == 'FAILURE'


class test_apply_tasks:
    """Tests for tasks called via apply() method."""

    def test_apply_single_task_ids(self, manager):
        """Test that a single task called via apply() has correct IDs."""
        @manager.app.task(bind=True)
        def single_apply_task(self):
            return {
                'task_id': self.request.id,
                'parent_id': self.request.parent_id,
                'root_id': self.request.root_id,
            }

        result = single_apply_task.apply()
        data = result.get()

        # Single task should have no parent and root_id should equal task_id
        assert data['parent_id'] is None
        assert data['root_id'] == data['task_id']

    def test_apply_nested_parent_child_relationship(self, manager):
        """Test parent-child relationship when one task calls another via apply()."""

        @manager.app.task(bind=True)
        def grandchild_task(task_self):
            return {
                'task_id': task_self.request.id,
                'parent_id': task_self.request.parent_id,
                'root_id': task_self.request.root_id,
                'name': 'grandchild_task'
            }

        @manager.app.task(bind=True)
        def child_task(task_self):

            # Call grandchild task via apply()
            grandchild_data = grandchild_task.apply().get()
            return {
                'task_id': task_self.request.id,
                'parent_id': task_self.request.parent_id,
                'root_id': task_self.request.root_id,
                'name': 'child_task',
                'grandchild_data': grandchild_data
            }

        @manager.app.task(bind=True)
        def parent_task(task_self):
            # Call child task via apply()
            child_data = child_task.apply().get()
            parent_data = {
                'task_id': task_self.request.id,
                'parent_id': task_self.request.parent_id,
                'root_id': task_self.request.root_id,
                'name': 'parent_task',
                'child_data': child_data
            }
            return parent_data

        result = parent_task.apply()

        parent_data = result.get()
        child_data = parent_data['child_data']
        grandchild_data = child_data['grandchild_data']

        # Verify parent task
        assert parent_data['name'] == 'parent_task'
        assert parent_data['parent_id'] is None
        assert parent_data['root_id'] == parent_data['task_id']

        # Verify child task
        assert child_data['name'] == 'child_task'
        assert child_data['parent_id'] == parent_data['task_id']
        assert child_data['root_id'] == parent_data['task_id']

        # Verify grandchild task
        assert grandchild_data['name'] == 'grandchild_task'
        assert grandchild_data['parent_id'] == child_data['task_id']
        assert grandchild_data['root_id'] == parent_data['task_id']


class test_trace_log_arguments:
    args = "CUSTOM ARGS"
    kwargs = "CUSTOM KWARGS"

    def assert_trace_log(self, caplog, result, expected):
        # wait for logs from worker
        time.sleep(.01)

        records = [(r.name, r.levelno, r.msg, r.data["args"], r.data["kwargs"])
                   for r in caplog.records
                   if r.name in {'celery.worker.strategy', 'celery.app.trace'}
                   if r.data["id"] == result.task_id
                   ]
        assert records == [(*e, self.args, self.kwargs) for e in expected]

    def call_task_with_reprs(self, task):
        return task.set(argsrepr=self.args, kwargsrepr=self.kwargs).delay()

    @flaky
    def test_task_success(self, caplog):
        result = self.call_task_with_reprs(add.s(2, 2))
        value = result.get()
        assert value == 4
        assert result.successful() is True

        self.assert_trace_log(caplog, result, [
            ('celery.worker.strategy', logging.INFO,
             celery.app.trace.LOG_RECEIVED,
             ),
            ('celery.app.trace', logging.INFO,
             celery.app.trace.LOG_SUCCESS,
             ),
        ])

    @flaky
    def test_task_failed(self, caplog):
        result = self.call_task_with_reprs(fail.s(2, 2))
        with pytest.raises(ExpectedException):
            result.get(timeout=5)
        assert result.failed() is True

        self.assert_trace_log(caplog, result, [
            ('celery.worker.strategy', logging.INFO,
             celery.app.trace.LOG_RECEIVED,
             ),
            ('celery.app.trace', logging.ERROR,
             celery.app.trace.LOG_FAILURE,
             ),
        ])


class test_task_redis_result_backend:
    @pytest.fixture()
    def manager(self, manager):
        if not manager.app.conf.result_backend.startswith('redis'):
            raise pytest.skip('Requires redis result backend.')

        return manager

    def test_ignoring_result_no_subscriptions(self, manager):
        channels_before_test = get_active_redis_channels()

        result = add_ignore_result.delay(1, 2)
        assert result.ignored is True

        new_channels = [channel for channel in get_active_redis_channels() if channel not in channels_before_test]
        assert new_channels == []

    @flaky
    def test_asyncresult_forget_cancels_subscription(self, manager):
        channels_before_test = get_active_redis_channels()

        result = add.delay(1, 2)
        assert set(get_active_redis_channels()) == {
            f"celery-task-meta-{result.id}".encode(), *channels_before_test
        }
        result.forget()

        new_channels = [channel for channel in get_active_redis_channels() if channel not in channels_before_test]
        assert new_channels == []

    @flaky
    def test_asyncresult_get_cancels_subscription(self, manager):
        channels_before_test = get_active_redis_channels()

        result = add.delay(1, 2)
        assert set(get_active_redis_channels()) == {
            f"celery-task-meta-{result.id}".encode(), *channels_before_test
        }
        assert result.get(timeout=3) == 3

        new_channels = [channel for channel in get_active_redis_channels() if channel not in channels_before_test]
        assert new_channels == []


class test_task_replacement:
    def test_replaced_task_nesting_level_0(self, manager):
        @task_received.connect
        def task_received_handler(request, **kwargs):
            nonlocal assertion_result

            try:
                # This tests mainly that the field even exists and set to default 0
                assertion_result = request.replaced_task_nesting < 1
            except Exception:
                assertion_result = False

        non_replaced_task = add.si(4, 2)
        res = non_replaced_task.delay()
        assertion_result = False
        assert res.get(timeout=TIMEOUT) == 6
        assert assertion_result

    def test_replaced_task_nesting_level_1(self, manager):
        if not manager.app.conf.result_backend.startswith("redis"):
            raise pytest.skip("Requires redis result backend.")

        redis_connection = get_redis_connection()
        redis_connection.delete("redis-echo")

        @task_received.connect
        def task_received_handler(request, **kwargs):
            nonlocal assertion_result

            try:
                assertion_result = request.replaced_task_nesting <= 2
            except Exception:
                assertion_result = False

        replaced_task = second_order_replace1.si()
        res = replaced_task.delay()
        assertion_result = False
        res.get(timeout=TIMEOUT)
        assert assertion_result
        redis_messages = list(redis_connection.lrange("redis-echo", 0, -1))
        expected_messages = [b"In A", b"In B", b"In/Out C", b"Out B", b"Out A"]
        assert redis_messages == expected_messages

    def test_replaced_task_nesting_chain(self, manager):
        if not manager.app.conf.result_backend.startswith("redis"):
            raise pytest.skip("Requires redis result backend.")

        redis_connection = get_redis_connection()
        redis_connection.delete("redis-echo")

        @task_received.connect
        def task_received_handler(request, **kwargs):
            nonlocal assertion_result

            try:
                assertion_result = request.replaced_task_nesting <= 3
            except Exception:
                assertion_result = False

        assertion_result = False
        chain_task = second_order_replace1.si() | add.si(4, 2)
        res = chain_task.delay()
        res.get(timeout=TIMEOUT)
        assert assertion_result
        redis_messages = list(redis_connection.lrange("redis-echo", 0, -1))
        expected_messages = [b"In A", b"In B", b"In/Out C", b"Out B", b"Out A"]
        assert redis_messages == expected_messages



import pytest
from pytest_celery import (RABBITMQ_CONTAINER_TIMEOUT, RESULT_TIMEOUT, CeleryBrokerCluster, CeleryTestSetup,
                           RabbitMQContainer, RabbitMQTestBroker)
from pytest_docker_tools import container, fxtr

from t.integration.tasks import identity

failover_broker = container(
    image="{default_rabbitmq_broker_image}",
    ports=fxtr("default_rabbitmq_broker_ports"),
    environment=fxtr("default_rabbitmq_broker_env"),
    network="{default_pytest_celery_network.name}",
    wrapper_class=RabbitMQContainer,
    timeout=RABBITMQ_CONTAINER_TIMEOUT,
)


@pytest.fixture
def failover_rabbitmq_broker(failover_broker: RabbitMQContainer) -> RabbitMQTestBroker:
    broker = RabbitMQTestBroker(failover_broker)
    yield broker
    broker.teardown()


@pytest.fixture
def celery_broker_cluster(
    celery_rabbitmq_broker: RabbitMQTestBroker,
    failover_rabbitmq_broker: RabbitMQTestBroker,
) -> CeleryBrokerCluster:
    cluster = CeleryBrokerCluster(celery_rabbitmq_broker, failover_rabbitmq_broker)
    yield cluster
    cluster.teardown()


class test_broker_failover:
    def test_killing_first_broker(self, celery_setup: CeleryTestSetup):
        assert len(celery_setup.broker_cluster) > 1
        celery_setup.broker.kill()
        expected = "test_broker_failover"
        res = identity.s(expected).apply_async(queue=celery_setup.worker.worker_queue)
        assert res.get(timeout=RESULT_TIMEOUT) == expected

    def test_reconnect_to_main(self, celery_setup: CeleryTestSetup):
        assert len(celery_setup.broker_cluster) > 1
        celery_setup.broker_cluster[0].kill()
        expected = "test_broker_failover"
        res = identity.s(expected).apply_async(queue=celery_setup.worker.worker_queue)
        assert res.get(timeout=RESULT_TIMEOUT) == expected
        celery_setup.broker_cluster[1].kill()
        celery_setup.broker_cluster[0].restart()
        res = identity.s(expected).apply_async(queue=celery_setup.worker.worker_queue)
        assert res.get(timeout=RESULT_TIMEOUT) == expected

    def test_broker_failover_ui(self, celery_setup: CeleryTestSetup):
        assert len(celery_setup.broker_cluster) > 1
        celery_setup.broker_cluster[0].kill()
        celery_setup.worker.assert_log_exists("Will retry using next failover.")
        celery_setup.worker.assert_log_exists(
            f"Connected to amqp://guest:**@{celery_setup.broker_cluster[1].hostname()}:5672//"
        )



from celery import Celery
from celery.worker.control import control_command, inspect_command


@control_command(
    args=[('a', int), ('b', int)],
    signature='a b',
)
def custom_control_cmd(state, a, b):
    """Ask the workers to reply with a and b."""
    return {'ok': f'Received {a} and {b}'}


@inspect_command(
    args=[('x', int)],
    signature='x',
)
def custom_inspect_cmd(state, x):
    """Ask the workers to reply with x."""
    return {'ok': f'Received {x}'}


app = Celery(set_as_current=False)
app.config_from_object('t.integration.test_worker_config')



"""
Integration tests for memory leak issue #8882.

These tests reproduce memory leak scenarios that occur when Celery tasks
raise unhandled exceptions, causing ExceptionInfo objects to not be
properly garbage collected.
"""

import gc
import logging
import os
import tracemalloc

from celery import Celery

logger = logging.getLogger(__name__)


class MemoryLeakUnhandledExceptionsTest:
    """Test class for memory leak scenarios with unhandled exceptions."""

    def __init__(self):
        self.app = Celery('test_memory_leak')
        self.app.conf.update(
            broker_url='memory://',
            result_backend='cache+memory://',
            task_always_eager=True,
            task_eager_propagates=True,
            task_store_eager_result=True,
        )
        self.setup_tasks()

    def setup_tasks(self):
        """Setup test tasks."""

        @self.app.task
        def task_success():
            """Task that completes successfully - baseline for memory comparison."""
            return "success"

        @self.app.task
        def task_unhandled_exception():
            """Task that raises an unhandled RuntimeError exception."""
            raise RuntimeError("Unhandled exception for memory leak test")

        @self.app.task(bind=True, max_retries=3)
        def task_retry_then_fail(self):
            """Task that retries multiple times and eventually fails with unhandled exception."""
            if self.request.retries < self.max_retries:
                raise self.retry(countdown=0.001)
            raise RuntimeError("Final retry failure - unhandled exception")

        @self.app.task
        def task_nested_exception_stack():
            """Task that raises exception through deeply nested function calls."""
            def deep_level_5():
                local_data = {"level": 5, "data": list(range(100))}  # noqa: F841
                raise ValueError("Deep nested exception at level 5")

            def deep_level_4():
                local_data = {"level": 4, "nested": {"data": list(range(50))}}  # noqa: F841
                deep_level_5()

            def deep_level_3():
                local_data = [1, 2, 3, {"nested": True}]  # noqa: F841
                deep_level_4()

            def deep_level_2():
                deep_level_3()

            def deep_level_1():
                deep_level_2()

            deep_level_1()

        self.task_success = task_success
        self.task_unhandled_exception = task_unhandled_exception
        self.task_retry_then_fail = task_retry_then_fail
        self.task_nested_exception_stack = task_nested_exception_stack


def get_memory_usage():
    """
    Get current memory usage in bytes.

    Returns RSS (total process memory) if psutil is available,
    otherwise returns Python heap allocations via tracemalloc.
    Note: These measurements are not directly comparable.
    """
    try:
        import psutil
        process = psutil.Process(os.getpid())
        return process.memory_info().rss
    except ImportError:
        # Fallback to tracemalloc if psutil not available
        current, peak = tracemalloc.get_traced_memory()
        return current


def test_mem_leak_unhandled_exceptions():
    """Test that reproduces the memory leak when tasks raise unhandled exceptions."""

    # Setup
    test_instance = MemoryLeakUnhandledExceptionsTest()

    # Enable memory tracing
    tracemalloc.start()

    # Warm up - run some successful tasks first
    for _ in range(50):
        try:
            test_instance.task_success.apply()
        except Exception:
            pass

    # Force garbage collection and get baseline memory
    gc.collect()
    baseline_memory = get_memory_usage()

    # Run many failing tasks - this should demonstrate the leak
    exception_count = 0
    for _ in range(500):  # Reduced from 1000 to make test faster
        try:
            test_instance.task_unhandled_exception.apply()
        except Exception:
            exception_count += 1

    # Force garbage collection
    gc.collect()
    after_exceptions_memory = get_memory_usage()

    # Run successful tasks again to ensure the leak is from exceptions
    for _ in range(50):
        try:
            test_instance.task_success.apply()
        except Exception:
            pass

    gc.collect()
    final_memory = get_memory_usage()

    # Calculate memory increase
    memory_increase = after_exceptions_memory - baseline_memory

    # Stop tracing
    tracemalloc.stop()

    # Log memory statistics for debugging
    logger.debug("--- Memory Statistics ---")  # Separator for better readability
    logger.debug(f"Baseline memory: {baseline_memory / 1024 / 1024:.2f} MB")
    logger.debug(f"After exceptions: {after_exceptions_memory / 1024 / 1024:.2f} MB")
    logger.debug(f"Final memory: {final_memory / 1024 / 1024:.2f} MB")
    logger.debug(f"Memory increase: {memory_increase / 1024 / 1024:.2f} MB")
    logger.debug(f"Exceptions processed: {exception_count}")

    # The test should demonstrate a significant memory increase
    # This threshold may need adjustment based on the system
    memory_increase_mb = memory_increase / 1024 / 1024

    # Verify the memory leak is fixed - memory increase should be minimal
    # Before fix: >70MB for 1000 tasks (~70KB/task)
    # After fix: <5MB for 500 tasks (<10KB/task)
    threshold_percent = float(os.getenv("MEMORY_LEAK_THRESHOLD_PERCENT", 10))  # Default: 10% increase
    memory_threshold_mb = baseline_memory / 1024 / 1024 * (threshold_percent / 100)
    assert memory_increase_mb < memory_threshold_mb, (
        f"Memory leak still exists! Expected <{memory_threshold_mb:.2f}MB increase "
        f"(based on {threshold_percent}% of baseline), "
        f"but got {memory_increase_mb:.2f}MB. "
        f"This indicates the memory leak fix is not working properly."
    )


def test_mem_leak_retry_failures():
    """Test memory leak with task retry and eventual failure scenarios."""

    test_instance = MemoryLeakUnhandledExceptionsTest()

    # Enable memory tracing
    tracemalloc.start()

    # Get baseline
    gc.collect()
    baseline_memory = get_memory_usage()

    # Run tasks that retry and eventually fail
    for _ in range(100):  # Fewer iterations since retries are expensive
        try:
            test_instance.task_retry_then_fail.apply()
        except Exception:
            pass

    gc.collect()
    after_retries_memory = get_memory_usage()

    # Stop tracing
    tracemalloc.stop()

    # Calculate memory increase
    memory_increase = after_retries_memory - baseline_memory
    memory_increase_mb = memory_increase / 1024 / 1024

    logger.debug("")  # New line for better readability
    logger.debug(f"Baseline memory: {baseline_memory / 1024 / 1024:.2f} MB")
    logger.debug(f"After retries: {after_retries_memory / 1024 / 1024:.2f} MB")
    logger.debug(f"Memory increase: {memory_increase_mb:.2f} MB")

    # Retries should not show significant memory increase if fix is working
    assert memory_increase_mb < 3, (
        f"Memory leak in retry scenarios! Expected <3MB increase for 100 retry tasks, "
        f"but got {memory_increase_mb:.2f}MB"
    )


def test_mem_leak_nested_exception_stacks():
    """Test memory leak with deeply nested exception stacks and local variables."""

    test_instance = MemoryLeakUnhandledExceptionsTest()

    # Enable memory tracing
    tracemalloc.start()

    # Get baseline
    gc.collect()
    baseline_memory = get_memory_usage()

    # Run tasks with complex exception stacks
    for _ in range(200):
        try:
            test_instance.task_nested_exception_stack.apply()
        except Exception:
            pass

    gc.collect()
    after_complex_memory = get_memory_usage()

    # Stop tracing
    tracemalloc.stop()

    # Calculate memory increase
    memory_increase = after_complex_memory - baseline_memory
    memory_increase_mb = memory_increase / 1024 / 1024

    logger.debug("Memory usage results:")
    logger.debug(f"Baseline memory: {baseline_memory / 1024 / 1024:.2f} MB")
    logger.debug(f"After complex exceptions: {after_complex_memory / 1024 / 1024:.2f} MB")
    logger.debug(f"Memory increase: {memory_increase_mb:.2f} MB")

    # Complex exceptions should not show significant memory increase if fix is working
    assert memory_increase_mb < 4, (
        f"Memory leak in nested exception scenarios! Expected <4MB increase for 200 nested tasks, "
        f"but got {memory_increase_mb:.2f}MB"
    )


if __name__ == "__main__":
    # Allow running these tests standalone for debugging
    print("Running memory leak integration tests...")
    test_mem_leak_unhandled_exceptions()
    test_mem_leak_retry_failures()
    test_mem_leak_nested_exception_stacks()
    print("Memory leak integration tests completed")



from unittest.mock import Mock

import pytest

from celery.backends.consul import ConsulBackend

pytest.importorskip('consul')


class test_ConsulBackend:

    def setup_method(self):
        self.backend = ConsulBackend(
            app=self.app, url='consul://localhost:800')

    def test_supports_autoexpire(self):
        assert self.backend.supports_autoexpire

    def test_consul_consistency(self):
        assert self.backend.consistency == 'consistent'

    def test_get(self):
        index = 100
        data = {'Key': 'test-consul-1', 'Value': 'mypayload'}
        self.backend.one_client = Mock(name='c.client')
        self.backend.one_client.kv.get.return_value = (index, data)
        assert self.backend.get(data['Key']) == 'mypayload'

    def test_set(self):
        self.backend.one_client = Mock(name='c.client')
        self.backend.one_client.session.create.return_value = 'c8dfa770-4ea3-2ee9-d141-98cf0bfe9c59'
        self.backend.one_client.kv.put.return_value = True
        assert self.backend.set('Key', 'Value') is True

    def test_delete(self):
        self.backend.one_client = Mock(name='c.client')
        self.backend.one_client.kv.delete.return_value = True
        assert self.backend.delete('Key') is True

    def test_index_bytes_key(self):
        key = 'test-consul-2'
        assert self.backend._key_to_consul_key(key) == key
        assert self.backend._key_to_consul_key(key.encode('utf-8')) == key



import os
from collections.abc import Iterable
from time import sleep

from pydantic import BaseModel

from celery import Signature, Task, chain, chord, group, shared_task
from celery.canvas import signature
from celery.exceptions import Reject, SoftTimeLimitExceeded
from celery.utils.log import get_task_logger

LEGACY_TASKS_DISABLED = True
try:
    # Imports that are not available in Celery 4
    from celery.canvas import StampingVisitor
except ImportError:
    LEGACY_TASKS_DISABLED = False


def get_redis_connection():
    from redis import StrictRedis

    host = os.environ.get("REDIS_HOST", "localhost")
    port = os.environ.get("REDIS_PORT", 6379)
    return StrictRedis(host=host, port=port)


logger = get_task_logger(__name__)


@shared_task
def identity(x):
    """Return the argument."""
    return x


@shared_task
def add(x, y, z=None):
    """Add two or three numbers."""
    if z:
        return x + y + z
    else:
        return x + y


@shared_task
def mul(x: int, y: int) -> int:
    """Multiply two numbers"""
    return x * y


@shared_task
def write_to_file_and_return_int(file_name, i):
    with open(file_name, mode='a', buffering=1) as file_handle:
        file_handle.write(str(i)+'\n')

    return i


@shared_task(typing=False)
def add_not_typed(x, y):
    """Add two numbers, but don't check arguments"""
    return x + y


@shared_task(ignore_result=True)
def add_ignore_result(x, y):
    """Add two numbers."""
    return x + y


@shared_task
def raise_error(*args):
    """Deliberately raise an error."""
    raise ValueError("deliberate error")


@shared_task
def chain_add(x, y):
    (
        add.s(x, x) | add.s(y)
    ).apply_async()


@shared_task
def chord_add(x, y):
    chord(add.s(x, x), add.s(y)).apply_async()


@shared_task
def delayed_sum(numbers, pause_time=1):
    """Sum the iterable of numbers."""
    # Allow the task to be in STARTED state for
    # a limited period of time.
    sleep(pause_time)
    return sum(numbers)


@shared_task
def delayed_sum_with_soft_guard(numbers, pause_time=1):
    """Sum the iterable of numbers."""
    try:
        sleep(pause_time)
        return sum(numbers)
    except SoftTimeLimitExceeded:
        return 0


@shared_task
def tsum(nums):
    """Sum an iterable of numbers."""
    return sum(nums)


@shared_task
def xsum(nums):
    """Sum of ints and lists."""
    return sum(sum(num) if isinstance(num, Iterable) else num for num in nums)


@shared_task(bind=True)
def add_replaced(self, x, y):
    """Add two numbers (via the add task)."""
    raise self.replace(add.s(x, y))


@shared_task(bind=True)
def replace_with_chain(self, *args, link_msg=None):
    c = chain(identity.s(*args), identity.s())
    link_sig = redis_echo.s()
    if link_msg is not None:
        link_sig.args = (link_msg,)
        link_sig.set(immutable=True)
    c.link(link_sig)

    return self.replace(c)


@shared_task(bind=True)
def replace_with_chain_which_raises(self, *args, link_msg=None):
    c = chain(identity.s(*args), raise_error.s())
    link_sig = redis_echo.s()
    if link_msg is not None:
        link_sig.args = (link_msg,)
        link_sig.set(immutable=True)
    c.link_error(link_sig)

    return self.replace(c)


@shared_task(bind=True)
def replace_with_empty_chain(self, *_):
    return self.replace(chain())


@shared_task(bind=True)
def add_to_all(self, nums, val):
    """Add the given value to all supplied numbers."""
    subtasks = [add.s(num, val) for num in nums]
    raise self.replace(group(*subtasks))


@shared_task(bind=True)
def add_to_all_to_chord(self, nums, val):
    for num in nums:
        self.add_to_chord(add.s(num, val))
    return 0


@shared_task(bind=True)
def add_chord_to_chord(self, nums, val):
    subtasks = [add.s(num, val) for num in nums]
    self.add_to_chord(group(subtasks) | tsum.s())
    return 0


@shared_task
def print_unicode(log_message='håå®ƒ valmuefrø', print_message='hiöäüß'):
    """Task that both logs and print strings containing funny characters."""
    logger.warning(log_message)
    print(print_message)


@shared_task
def return_exception(e):
    """Return a tuple containing the exception message and sentinel value."""
    return e, True


@shared_task
def sleeping(i, **_):
    """Task sleeping for ``i`` seconds, and returning nothing."""
    sleep(i)


@shared_task(bind=True)
def ids(self, i):
    """Returns a tuple of ``root_id``, ``parent_id`` and
    the argument passed as ``i``."""
    return self.request.root_id, self.request.parent_id, i


@shared_task(bind=True)
def collect_ids(self, res, i):
    """Used as a callback in a chain or group where the previous tasks
    are :task:`ids`: returns a tuple of::

        (previous_result, (root_id, parent_id, i))
    """
    return res, (self.request.root_id, self.request.parent_id, i)


@shared_task(bind=True, default_retry_delay=1)
def retry(self, return_value=None):
    """Task simulating multiple retries.

    When return_value is provided, the task after retries returns
    the result. Otherwise it fails.
    """
    if return_value:
        attempt = getattr(self, 'attempt', 0)
        print('attempt', attempt)
        if attempt >= 3:
            delattr(self, 'attempt')
            return return_value
        self.attempt = attempt + 1

    raise self.retry(exc=ExpectedException(), countdown=5)


@shared_task(bind=True, default_retry_delay=1)
def retry_unpickleable(self, foo, bar, *, retry_kwargs):
    """Task that fails with an unpickleable exception and is retried."""
    raise self.retry(exc=UnpickleableException(foo, bar), **retry_kwargs)


@shared_task(bind=True, expires=120.0, max_retries=1)
def retry_once(self, *args, expires=None, max_retries=1, countdown=0.1):
    """Task that fails and is retried. Returns the number of retries."""
    if self.request.retries:
        return self.request.retries
    raise self.retry(countdown=countdown,
                     expires=expires,
                     max_retries=max_retries)


@shared_task(bind=True, max_retries=1)
def retry_once_priority(self, *args, expires=60.0, max_retries=1,
                        countdown=0.1):
    """Task that fails and is retried. Returns the priority."""
    if self.request.retries:
        return self.request.delivery_info['priority']
    raise self.retry(countdown=countdown,
                     max_retries=max_retries)


@shared_task(bind=True, max_retries=1)
def retry_once_headers(self, *args, max_retries=1,
                       countdown=0.1):
    """Task that fails and is retried. Returns headers."""
    if self.request.retries:
        return self.request.headers
    raise self.retry(countdown=countdown,
                     max_retries=max_retries)


@shared_task
def redis_echo(message, redis_key="redis-echo"):
    """Task that appends the message to a redis list."""
    redis_connection = get_redis_connection()
    redis_connection.rpush(redis_key, message)


@shared_task(bind=True)
def redis_echo_group_id(self, _, redis_key="redis-group-ids"):
    redis_connection = get_redis_connection()
    redis_connection.rpush(redis_key, self.request.group)


@shared_task
def redis_count(redis_key="redis-count"):
    """Task that increments a specified or well-known redis key."""
    redis_connection = get_redis_connection()
    redis_connection.incr(redis_key)


@shared_task(bind=True)
def second_order_replace1(self, state=False):
    redis_connection = get_redis_connection()
    if not state:
        redis_connection.rpush('redis-echo', 'In A')
        new_task = chain(second_order_replace2.s(),
                         second_order_replace1.si(state=True))
        raise self.replace(new_task)
    else:
        redis_connection.rpush('redis-echo', 'Out A')


@shared_task(bind=True)
def second_order_replace2(self, state=False):
    redis_connection = get_redis_connection()
    if not state:
        redis_connection.rpush('redis-echo', 'In B')
        new_task = chain(redis_echo.s("In/Out C"),
                         second_order_replace2.si(state=True))
        raise self.replace(new_task)
    else:
        redis_connection.rpush('redis-echo', 'Out B')


@shared_task(bind=True)
def build_chain_inside_task(self):
    """Task to build a chain.

    This task builds a chain and returns the chain's AsyncResult
    to verify that Asyncresults are correctly converted into
    serializable objects"""
    test_chain = (
        add.s(1, 1) |
        add.s(2) |
        group(
            add.s(3),
            add.s(4)
        ) |
        add.s(5)
    )
    result = test_chain()
    return result


class ExpectedException(Exception):
    """Sentinel exception for tests."""

    def __eq__(self, other):
        return (
            other is not None and
            isinstance(other, ExpectedException) and
            self.args == other.args
        )

    def __hash__(self):
        return hash(self.args)


class UnpickleableException(Exception):
    """Exception that doesn't survive a pickling roundtrip (dump + load)."""

    def __init__(self, foo, bar=None):
        if bar is None:
            # We define bar with a default value in the signature so that
            # it's easier to add a break point here to find out when the
            # exception is being unpickled.
            raise TypeError("bar must be provided")

        super().__init__(foo)
        self.bar = bar


@shared_task
def fail(*args):
    """Task that simply raises ExpectedException."""
    args = ("Task expected to fail",) + args
    raise ExpectedException(*args)


@shared_task()
def fail_unpickleable(foo, bar):
    """Task that raises an unpickleable exception."""
    raise UnpickleableException(foo, bar)


@shared_task(bind=True)
def fail_replaced(self, *args):
    """Replace this task with one which raises ExpectedException."""
    raise self.replace(fail.si(*args))


@shared_task(bind=True)
def return_priority(self, *_args):
    return "Priority: %s" % self.request.delivery_info['priority']


@shared_task(bind=True)
def return_properties(self):
    return self.request.properties


class ClassBasedAutoRetryTask(Task):
    name = 'auto_retry_class_task'
    autoretry_for = (ValueError,)
    retry_kwargs = {'max_retries': 1}
    retry_backoff = True

    def run(self):
        if self.request.retries:
            return self.request.retries
        raise ValueError()


# The signatures returned by these tasks wouldn't actually run because the
# arguments wouldn't be fulfilled - we never actually delay them so it's fine
@shared_task
def return_nested_signature_chain_chain():
    return chain(chain([add.s()]))


@shared_task
def return_nested_signature_chain_group():
    return chain(group([add.s()]))


@shared_task
def return_nested_signature_chain_chord():
    return chain(chord([add.s()], add.s()))


@shared_task
def return_nested_signature_group_chain():
    return group(chain([add.s()]))


@shared_task
def return_nested_signature_group_group():
    return group(group([add.s()]))


@shared_task
def return_nested_signature_group_chord():
    return group(chord([add.s()], add.s()))


@shared_task
def return_nested_signature_chord_chain():
    return chord(chain([add.s()]), add.s())


@shared_task
def return_nested_signature_chord_group():
    return chord(group([add.s()]), add.s())


@shared_task
def return_nested_signature_chord_chord():
    return chord(chord([add.s()], add.s()), add.s())


@shared_task
def rebuild_signature(sig_dict):
    sig_obj = Signature.from_dict(sig_dict)

    def _recurse(sig):
        if not isinstance(sig, Signature):
            raise TypeError(f"{sig!r} is not a signature object")
        # Most canvas types have a `tasks` attribute
        if isinstance(sig, (chain, group, chord)):
            for task in sig.tasks:
                _recurse(task)
        # `chord`s also have a `body` attribute
        if isinstance(sig, chord):
            _recurse(sig.body)
    _recurse(sig_obj)


@shared_task
def errback_old_style(request_id):
    redis_count(request_id)
    return request_id


@shared_task
def errback_new_style(request, exc, tb):
    redis_count(request.id)
    return request.id


@shared_task
def replaced_with_me():
    return True


class AddParameterModel(BaseModel):
    x: int
    y: int


class AddResultModel(BaseModel):
    result: int


@shared_task(pydantic=True)
def add_pydantic(data: AddParameterModel) -> AddResultModel:
    """Add two numbers, but with parameters and results using Pydantic model serialization."""
    value = data.x + data.y
    return AddResultModel(result=value)


@shared_task(pydantic=True)
def add_pydantic_string_annotations(data: "AddParameterModel") -> "AddResultModel":
    """Add two numbers, but with string-annotated Pydantic models (__future__.annotations bug)."""
    value = data.x + data.y
    return AddResultModel(result=value)


if LEGACY_TASKS_DISABLED:
    class StampOnReplace(StampingVisitor):
        stamp = {"StampOnReplace": "This is the replaced task"}

        def on_signature(self, sig, **headers) -> dict:
            return self.stamp

    class StampedTaskOnReplace(Task):
        """Custom task for stamping on replace"""

        def on_replace(self, sig):
            sig.stamp(StampOnReplace())
            return super().on_replace(sig)

    @shared_task(bind=True, base=StampedTaskOnReplace)
    def replace_with_stamped_task(self: StampedTaskOnReplace, replace_with=None):
        if replace_with is None:
            replace_with = replaced_with_me.s()
        self.replace(signature(replace_with))


@shared_task(bind=True, acks_late=True)
def store_success_then_reject(self):
    """First delivery: store SUCCESS manually, then Reject to trigger redelivery.
    Second delivery: dedup finds SUCCESS, dispatches chain."""
    from celery.backends.base import states
    if not self.request.delivery_info.get('redelivered'):
        self.backend.store_result(self.request.id, 'first-pass', states.SUCCESS)
        raise Reject(requeue=True)
    # When dedup is enabled the fast-path intercepts before reaching here,
    # so 'dedup-pass' is only returned when dedup is disabled.
    return 'dedup-pass'


@shared_task(bind=True, acks_late=True)
def reject_then_succeed(self):
    """First delivery: Reject(requeue=True). Second delivery: succeed normally."""
    if not self.request.delivery_info.get('redelivered'):
        raise Reject(requeue=True)
    return 'second-pass'


@shared_task(soft_time_limit=2, time_limit=1)
def soft_time_limit_must_exceed_time_limit():
    pass



from unittest.mock import Mock, call, patch

import pytest

from celery import states
from celery.backends import azureblockblob
from celery.backends.azureblockblob import AzureBlockBlobBackend
from celery.exceptions import ImproperlyConfigured

MODULE_TO_MOCK = "celery.backends.azureblockblob"

pytest.importorskip('azure.storage.blob')
pytest.importorskip('azure.core.exceptions')


class test_AzureBlockBlobBackend:
    def setup_method(self):
        self.url = (
            "azureblockblob://"
            "DefaultEndpointsProtocol=protocol;"
            "AccountName=name;"
            "AccountKey=key;"
            "EndpointSuffix=suffix")

        self.backend = AzureBlockBlobBackend(
            app=self.app,
            url=self.url)

    @pytest.fixture(params=['', 'my_folder/'])
    def base_path(self, request):
        return request.param

    def test_missing_third_party_sdk(self):
        azurestorage = azureblockblob.azurestorage
        try:
            azureblockblob.azurestorage = None
            with pytest.raises(ImproperlyConfigured):
                AzureBlockBlobBackend(app=self.app, url=self.url)
        finally:
            azureblockblob.azurestorage = azurestorage

    def test_bad_connection_url(self):
        with pytest.raises(ImproperlyConfigured):
            AzureBlockBlobBackend._parse_url("azureblockblob://")

        with pytest.raises(ImproperlyConfigured):
            AzureBlockBlobBackend._parse_url("")

    @patch(MODULE_TO_MOCK + ".BlobServiceClient")
    def test_create_client(self, mock_blob_service_factory):
        mock_blob_service_client_instance = Mock()
        mock_blob_service_factory.from_connection_string.return_value = mock_blob_service_client_instance
        backend = AzureBlockBlobBackend(app=self.app, url=self.url)

        # ensure container gets created on client access...
        assert mock_blob_service_client_instance.create_container.call_count == 0
        assert backend._blob_service_client is not None
        assert mock_blob_service_client_instance.create_container.call_count == 1

        # ...but only once per backend instance
        assert backend._blob_service_client is not None
        assert mock_blob_service_client_instance.create_container.call_count == 1

    @patch(MODULE_TO_MOCK + ".AzureStorageQueuesTransport")
    @patch(MODULE_TO_MOCK + ".BlobServiceClient")
    def test_create_client__default_azure_credentials(self, mock_blob_service_client, mock_kombu_transport):
        credential_mock = Mock()
        mock_blob_service_client.return_value = Mock()
        mock_kombu_transport.parse_uri.return_value = (credential_mock, "dummy_account_url")
        url = "azureblockblob://DefaultAzureCredential@dummy_account_url"
        backend = AzureBlockBlobBackend(app=self.app, url=url)
        assert backend._blob_service_client is not None
        mock_kombu_transport.parse_uri.assert_called_once_with(url.replace("azureblockblob://", ""))
        mock_blob_service_client.assert_called_once_with(
            account_url="dummy_account_url",
            credential=credential_mock,
            connection_timeout=backend._connection_timeout,
            read_timeout=backend._read_timeout,
        )

    @patch(MODULE_TO_MOCK + ".AzureStorageQueuesTransport")
    @patch(MODULE_TO_MOCK + ".BlobServiceClient")
    def test_create_client__managed_identity_azure_credentials(self, mock_blob_service_client, mock_kombu_transport):
        credential_mock = Mock()
        mock_blob_service_client.return_value = Mock()
        mock_kombu_transport.parse_uri.return_value = (credential_mock, "dummy_account_url")
        url = "azureblockblob://ManagedIdentityCredential@dummy_account_url"
        backend = AzureBlockBlobBackend(app=self.app, url=url)
        assert backend._blob_service_client is not None
        mock_kombu_transport.parse_uri.assert_called_once_with(url.replace("azureblockblob://", ""))
        mock_blob_service_client.assert_called_once_with(
            account_url="dummy_account_url",
            credential=credential_mock,
            connection_timeout=backend._connection_timeout,
            read_timeout=backend._read_timeout,
        )

    @patch(MODULE_TO_MOCK + ".BlobServiceClient")
    def test_configure_client(self, mock_blob_service_factory):

        connection_timeout = 3
        read_timeout = 11
        self.app.conf.update(
            {
                'azureblockblob_connection_timeout': connection_timeout,
                'azureblockblob_read_timeout': read_timeout,
            }
        )

        mock_blob_service_client_instance = Mock()
        mock_blob_service_factory.from_connection_string.return_value = (
            mock_blob_service_client_instance
        )

        base_url = "azureblockblob://"
        connection_string = "connection_string"
        backend = AzureBlockBlobBackend(
            app=self.app, url=f'{base_url}{connection_string}'
        )

        client = backend._blob_service_client
        assert client is mock_blob_service_client_instance

        (
            mock_blob_service_factory
            .from_connection_string
            .assert_called_once_with(
                connection_string,
                connection_timeout=connection_timeout,
                read_timeout=read_timeout
            )
        )

    @patch(MODULE_TO_MOCK + ".AzureBlockBlobBackend._blob_service_client")
    def test_get(self, mock_client, base_path):
        self.backend.base_path = base_path
        self.backend.get(b"mykey")

        mock_client.get_blob_client \
            .assert_called_once_with(blob=base_path + "mykey", container="celery")

        mock_client.get_blob_client.return_value \
            .download_blob.return_value \
            .readall.return_value \
            .decode.assert_called_once()

    @patch(MODULE_TO_MOCK + ".AzureBlockBlobBackend._blob_service_client")
    def test_get_missing(self, mock_client):
        mock_client.get_blob_client.return_value \
            .download_blob.return_value \
            .readall.side_effect = azureblockblob.ResourceNotFoundError

        assert self.backend.get(b"mykey") is None

    @patch(MODULE_TO_MOCK + ".AzureBlockBlobBackend._blob_service_client")
    def test_set(self, mock_client, base_path):
        self.backend.base_path = base_path
        self.backend._set_with_state(b"mykey", "myvalue", states.SUCCESS)

        mock_client.get_blob_client.assert_called_once_with(
            container="celery", blob=base_path + "mykey")

        mock_client.get_blob_client.return_value \
            .upload_blob.assert_called_once_with("myvalue", overwrite=True)

    @patch(MODULE_TO_MOCK + ".AzureBlockBlobBackend._blob_service_client")
    def test_mget(self, mock_client, base_path):
        keys = [b"mykey1", b"mykey2"]

        self.backend.base_path = base_path
        self.backend.mget(keys)

        mock_client.get_blob_client.assert_has_calls(
            [call(blob=base_path + key.decode(), container='celery') for key in keys],
            any_order=True,)

    @patch(MODULE_TO_MOCK + ".AzureBlockBlobBackend._blob_service_client")
    def test_delete(self, mock_client, base_path):
        self.backend.base_path = base_path
        self.backend.delete(b"mykey")

        mock_client.get_blob_client.assert_called_once_with(
            container="celery", blob=base_path + "mykey")

        mock_client.get_blob_client.return_value \
            .delete_blob.assert_called_once()

    def test_base_path_conf(self, base_path):
        self.app.conf.azureblockblob_base_path = base_path
        backend = AzureBlockBlobBackend(
            app=self.app,
            url=self.url
        )
        assert backend.base_path == base_path

    def test_base_path_conf_default(self):
        backend = AzureBlockBlobBackend(
            app=self.app,
            url=self.url
        )
        assert backend.base_path == ''


class test_as_uri:
    def setup_method(self):
        self.url = (
            "azureblockblob://"
            "DefaultEndpointsProtocol=protocol;"
            "AccountName=name;"
            "AccountKey=account_key;"
            "EndpointSuffix=suffix"
        )
        self.backend = AzureBlockBlobBackend(
            app=self.app,
            url=self.url
        )

    def test_as_uri_include_password(self):
        assert self.backend.as_uri(include_password=True) == self.url

    def test_as_uri_exclude_password(self):
        assert self.backend.as_uri(include_password=False) == (
            "azureblockblob://"
            "DefaultEndpointsProtocol=protocol;"
            "AccountName=name;"
            "AccountKey=**;"
            "EndpointSuffix=suffix"
        )



"""myapp.py

Usage::

   (window1)$ python myapp.py worker -l INFO

   (window2)$ celery shell
   >>> from myapp import example
   >>> example()


You can also specify the app to use with the `celery` command,
using the `-A` / `--app` option::

    $ celery -A myapp worker -l INFO

With the `-A myproj` argument the program will search for an app
instance in the module ``myproj``.  You can also specify an explicit
name using the fully qualified form::

    $ celery -A myapp:app worker -l INFO

"""

import os
from datetime import UTC, datetime, timedelta

from declare_queue import my_quorum_queue

from celery import Celery
from celery.canvas import group

app = Celery("myapp", broker="amqp://guest@localhost//")

# Use custom queue (Optional) or set the default queue type to "quorum"
# app.conf.task_queues = (my_quorum_queue,)  # uncomment to use custom queue
app.conf.task_default_queue_type = "quorum"  # comment to use classic queue

# Required by Quorum Queues: https://www.rabbitmq.com/docs/quorum-queues#use-cases
app.conf.broker_transport_options = {"confirm_publish": True}

# Reduce qos to 4 (Optional, useful for testing)
app.conf.worker_prefetch_multiplier = 1
app.conf.worker_concurrency = 4

# Reduce logs (Optional, useful for testing)
app.conf.worker_heartbeat = None
app.conf.broker_heartbeat = 0


def is_using_quorum_queues(app) -> bool:
    queues = app.amqp.queues
    for qname in queues:
        qarguments = queues[qname].queue_arguments or {}
        if qarguments.get("x-queue-type") == "quorum":
            return True

    return False


@app.task
def add(x, y):
    return x + y


@app.task
def identity(x):
    return x


def example():
    queue = my_quorum_queue.name if my_quorum_queue in (app.conf.task_queues or {}) else "celery"

    while True:
        print("Celery Quorum Queue Example")
        print("===========================")
        print("1. Send a simple identity task")
        print("1.1 Send an ETA identity task")
        print("2. Send a group of add tasks")
        print("3. Inspect the active queues")
        print("4. Shutdown Celery worker")
        print("Q. Quit")
        print("Q! Exit")
        choice = input("Enter your choice (1-4 or Q): ")

        if choice == "1" or choice == "1.1":
            queue_type = "Quorum" if is_using_quorum_queues(app) else "Classic"
            payload = f"Hello, {queue_type} Queue!"
            eta = datetime.now(UTC) + timedelta(seconds=30)
            if choice == "1.1":
                result = identity.si(payload).apply_async(queue=queue, eta=eta)
            else:
                result = identity.si(payload).apply_async(queue=queue)
            print()
            print(f"Task sent with ID: {result.id}")
            print("Task type: identity")

            if choice == "1.1":
                print(f"ETA: {eta}")

            print(f"Payload: {payload}")

        elif choice == "2":
            tasks = [
                (1, 2),
                (3, 4),
                (5, 6),
            ]
            result = group(
                add.s(*tasks[0]),
                add.s(*tasks[1]),
                add.s(*tasks[2]),
            ).apply_async(queue=queue)
            print()
            print("Group of tasks sent.")
            print(f"Group result ID: {result.id}")
            for i, task_args in enumerate(tasks, 1):
                print(f"Task {i} type: add")
                print(f"Payload: {task_args}")

        elif choice == "3":
            active_queues = app.control.inspect().active_queues()
            print()
            print("Active queues:")
            for worker, queues in active_queues.items():
                print(f"Worker: {worker}")
                for q in queues:
                    print(f"  - {q['name']}")

        elif choice == "4":
            print("Shutting down Celery worker...")
            app.control.shutdown()

        elif choice.lower() == "q":
            print("Quitting test()")
            break

        elif choice.lower() == "q!":
            print("Exiting...")
            os.abort()

        else:
            print("Invalid choice. Please enter a number between 1 and 4 or Q to quit.")

        print("\n" + "#" * 80 + "\n")


if __name__ == "__main__":
    app.start()



from contextlib import contextmanager
from unittest.mock import MagicMock, Mock, patch

import pytest

from celery.concurrency.thread import TaskPool as ThreadTaskPool
from celery.fixups.django import DjangoFixup, DjangoWorkerFixup, FixupWarning, _maybe_close_fd, fixup
from t.unit import conftest


class FixupCase:
    Fixup = None

    @contextmanager
    def fixup_context(self, app, **kwargs):
        with patch('celery.fixups.django.DjangoWorkerFixup.validate_models'):
            with patch('celery.fixups.django.symbol_by_name') as symbyname:
                with patch('celery.fixups.django.import_module') as impmod:
                    worker = Mock()
                    worker.pool_cls = Mock(__module__='celery.concurrency.prefork')
                    f = self.Fixup(app, **kwargs)
                    f.worker = worker
                    yield f, impmod, symbyname


class test_DjangoFixup(FixupCase):
    Fixup = DjangoFixup

    def test_setting_default_app(self):
        from celery import _state
        prev, _state.default_app = _state.default_app, None
        try:
            app = Mock(name='app')
            DjangoFixup(app)
            app.set_default.assert_called_with()
        finally:
            _state.default_app = prev

    @patch('celery.fixups.django.DjangoWorkerFixup')
    def test_worker_fixup_property(self, DjangoWorkerFixup):
        f = DjangoFixup(self.app)
        f._worker_fixup = None
        assert f.worker_fixup is DjangoWorkerFixup()
        assert f.worker_fixup is DjangoWorkerFixup()

    def test_on_import_modules(self):
        f = DjangoFixup(self.app)
        f.worker_fixup = Mock(name='worker_fixup')
        f.on_import_modules()
        f.worker_fixup.validate_models.assert_called_with()

    def test_autodiscover_tasks(self, patching):
        patching.modules('django.apps')
        from django.apps import apps
        f = DjangoFixup(self.app)
        configs = [Mock(name='c1'), Mock(name='c2')]
        apps.get_app_configs.return_value = configs
        assert f.autodiscover_tasks() == [c.name for c in configs]

    @pytest.mark.masked_modules('django')
    def test_fixup_no_django(self, patching, mask_modules):
        with patch('celery.fixups.django.DjangoFixup') as Fixup:
            patching.setenv('DJANGO_SETTINGS_MODULE', '')
            fixup(self.app)
            Fixup.assert_not_called()

            patching.setenv('DJANGO_SETTINGS_MODULE', 'settings')
            with pytest.warns(FixupWarning):
                fixup(self.app)
            Fixup.assert_not_called()

    def test_fixup(self, patching):
        with patch('celery.fixups.django.DjangoFixup') as Fixup:
            patching.setenv('DJANGO_SETTINGS_MODULE', '')
            fixup(self.app)
            Fixup.assert_not_called()

            patching.setenv('DJANGO_SETTINGS_MODULE', 'settings')
            with conftest.module_exists('django'):
                import django
                django.VERSION = (1, 11, 1)
                fixup(self.app)
                Fixup.assert_called()

    def test_maybe_close_fd(self):
        with patch('os.close'):
            _maybe_close_fd(Mock())
            _maybe_close_fd(object())

    def test_init(self):
        with self.fixup_context(self.app) as (f, importmod, sym):
            assert f

    @pytest.mark.patched_module(
        'django',
        'django.db',
        'django.db.transaction',
    )
    def test_install(self, patching, module):
        self.app.loader = Mock()
        self.cw = patching('os.getcwd')
        self.p = patching('sys.path')
        self.sigs = patching('celery.fixups.django.signals')
        with self.fixup_context(self.app) as (f, _, _):
            self.cw.return_value = '/opt/vandelay'
            f.install()
            self.sigs.worker_init.connect.assert_called_with(f.on_worker_init)
            assert self.app.loader.now == f.now

            # Specialized DjangoTask class is used
            assert self.app.task_cls == 'celery.contrib.django.task:DjangoTask'
            from celery.contrib.django.task import DjangoTask
            assert issubclass(f.app.Task, DjangoTask)
            assert hasattr(f.app.Task, 'delay_on_commit')
            assert hasattr(f.app.Task, 'apply_async_on_commit')

            self.p.insert.assert_called_with(0, '/opt/vandelay')

    def test_install_custom_user_task(self, patching):
        patching('celery.fixups.django.signals')

        self.app.task_cls = 'myapp.celery.tasks:Task'
        self.app._custom_task_cls_used = True

        with self.fixup_context(self.app) as (f, _, _):
            f.install()
            # Specialized DjangoTask class is NOT used,
            # The one from the user's class is
            assert self.app.task_cls == 'myapp.celery.tasks:Task'

    def test_install_custom_user_task_as_class_attribute(self, patching):
        patching('celery.fixups.django.signals')

        from celery.app import Celery

        class MyCeleryApp(Celery):
            task_cls = 'myapp.celery.tasks:Task'

        app = MyCeleryApp('mytestapp')

        with self.fixup_context(app) as (f, _, _):
            f.install()
            # Specialized DjangoTask class is NOT used,
            # The one from the user's class is
            assert app.task_cls == 'myapp.celery.tasks:Task'

    def test_now(self):
        with self.fixup_context(self.app) as (f, _, _):
            assert f.now(utc=True)
            f._now.assert_not_called()
            assert f.now(utc=False)
            f._now.assert_called()

    def test_on_worker_init(self):
        with self.fixup_context(self.app) as (f, _, _):
            with patch('celery.fixups.django.DjangoWorkerFixup') as DWF:
                mock_worker = Mock(name="worker")
                f.on_worker_init(sender=mock_worker)
                assert DWF.return_value.worker == mock_worker

                DWF.assert_called_with(f.app)
                DWF.return_value.install.assert_called_with()
                assert f._worker_fixup is DWF.return_value

    def test_on_worker_init_warns_without_sender(self):
        with self.fixup_context(self.app) as (f, _, _):
            with patch("celery.fixups.django.DjangoWorkerFixup"):
                with pytest.warns(FixupWarning, match="called without a sender"):
                    f.on_worker_init(sender=None)


class InterfaceError(Exception):
    pass


class test_DjangoWorkerFixup(FixupCase):
    Fixup = DjangoWorkerFixup

    def test_init(self):
        with self.fixup_context(self.app) as (f, importmod, sym):
            assert f

    def test_install(self):
        self.app.conf = {'CELERY_DB_REUSE_MAX': None}
        self.app.loader = Mock()
        with self.fixup_context(self.app) as (f, _, _):
            with patch('celery.fixups.django.signals') as sigs:
                f.install()
                sigs.beat_embedded_init.connect.assert_called_with(
                    f.close_database,
                )
                sigs.task_prerun.connect.assert_called_with(f.on_task_prerun)
                sigs.task_postrun.connect.assert_called_with(f.on_task_postrun)
                sigs.worker_process_init.connect.assert_called_with(
                    f.on_worker_process_init,
                )

    def test_on_worker_process_init(self, patching):
        with self.fixup_context(self.app) as (f, _, _):
            with patch('celery.fixups.django._maybe_close_fd', side_effect=InterfaceError) as mcf:
                _all = f._db.connections.all = Mock()
                conns = _all.return_value = [
                    Mock(), MagicMock(),
                ]
                conns[0].connection = None
                with patch.object(f, 'close_cache'):
                    with patch.object(f, '_close_database'):
                        f.interface_errors = (InterfaceError, )
                        f.on_worker_process_init()
                        mcf.assert_called_with(conns[1].connection)
                        f.close_cache.assert_called_with()
                        f._close_database.assert_called_with()

                        f.validate_models = Mock(name='validate_models')
                        patching.setenv('FORKED_BY_MULTIPROCESSING', '1')
                        f.on_worker_process_init()
                        f.validate_models.assert_called_with()

    def test_on_task_prerun(self):
        task = Mock()
        with self.fixup_context(self.app) as (f, _, _):
            task.request.is_eager = False
            with patch.object(f, 'close_database'):
                f.on_task_prerun(task)
                f.close_database.assert_called_with()

            task.request.is_eager = True
            with patch.object(f, 'close_database'):
                f.on_task_prerun(task)
                f.close_database.assert_not_called()

    def test_on_task_postrun(self):
        task = Mock()
        with self.fixup_context(self.app) as (f, _, _):
            with patch.object(f, 'close_cache'):
                task.request.is_eager = False
                with patch.object(f, 'close_database'):
                    f.on_task_postrun(task)
                    f.close_database.assert_called()
                    f.close_cache.assert_called()

            # when a task is eager, don't close connections
            with patch.object(f, 'close_cache'):
                task.request.is_eager = True
                with patch.object(f, 'close_database'):
                    f.on_task_postrun(task)
                    f.close_database.assert_not_called()
                    f.close_cache.assert_not_called()

    def test_close_database(self):
        with self.fixup_context(self.app) as (f, _, _):
            with patch.object(f, '_close_database') as _close:
                f.db_reuse_max = None
                f.close_database()
                _close.assert_called_with()
                _close.reset_mock()

                f.db_reuse_max = 10
                f._db_recycles = 3
                f.close_database()
                _close.assert_not_called()
                assert f._db_recycles == 4
                _close.reset_mock()

                f._db_recycles = 20
                f.close_database()
                _close.assert_called_with()
                assert f._db_recycles == 1

    def test__close_database(self):
        with self.fixup_context(self.app) as (f, _, _):
            conns = [Mock(), Mock(), Mock()]
            conns[1].close.side_effect = KeyError('already closed')
            f.DatabaseError = KeyError
            f.interface_errors = ()

            f._db.connections = Mock()  # ConnectionHandler
            f._db.connections.all.side_effect = lambda initialized_only: conns

            f._close_database()
            conns[0].close.assert_called_with()
            conns[1].close.assert_called_with()
            conns[2].close.assert_called_with()

            conns[1].close.side_effect = KeyError(
                'omg')
            with pytest.raises(KeyError):
                f._close_database()

    def test__close_database_django_pre_41(self):
        """Test that Django < 4.1 (without initialized_only parameter) is handled."""
        with self.fixup_context(self.app) as (f, _, _):
            conns = [Mock(), Mock()]
            f.DatabaseError = KeyError
            f.interface_errors = ()

            # Mock Django < 4.1 behavior: connections.all() doesn't accept initialized_only
            f._db.connections = Mock()

            def all_without_initialized_only(**kwargs):
                if 'initialized_only' in kwargs:
                    raise TypeError("all() got an unexpected keyword argument 'initialized_only'")
                return conns

            f._db.connections.all = Mock(side_effect=all_without_initialized_only)

            # Should fall back to calling all() without initialized_only
            f._close_database()

            # Verify it was called twice: first with initialized_only (raises), then without
            assert f._db.connections.all.call_count == 2
            # First call with initialized_only=True
            f._db.connections.all.assert_any_call(initialized_only=True)
            # Second call without arguments (fallback)
            f._db.connections.all.assert_any_call()

            # Verify connections were closed
            conns[0].close.assert_called_with()
            conns[1].close.assert_called_with()

    def test_close_database_always_closes_connections(self):
        with self.fixup_context(self.app) as (f, _, _):
            conn = Mock()
            f._db.connections.all = Mock(return_value=[conn])
            f.close_database()
            conn.close.assert_called_once_with()
            # close_if_unusable_or_obsolete is not safe to call in all conditions, so avoid using
            # it to optimize connection handling.
            conn.close_if_unusable_or_obsolete.assert_not_called()

    def test_close_database_skip_conn_pool(self):
        class Connection:
            """Mock connection without `close_pool` method."""
            alias = 'default'

            def close(self):
                pass

        with self.fixup_context(self.app) as (f, _, _):
            conn = Mock(spec=Connection)
            f._db.connections.all = Mock(return_value=[conn])
            f.close_database()
            assert not hasattr(conn, "close_pool")
            conn.close.assert_called_once_with()

    def test_close_database_suppresses_close_pool_keyerror(self):
        with self.fixup_context(self.app) as (f, _, _):
            conn = Mock()
            conn.close_pool = Mock(side_effect=KeyError("pool already closed"))
            f._db.connections.all = Mock(return_value=[conn])
            f.close_database()  # should not raise
            conn.close.assert_called_once_with()
            conn.close_pool.assert_called_once_with()

    def test_close_database_conn_pool_based_on_settings(self):
        class DJSettings:
            DATABASES = {}

        with self.fixup_context(self.app) as (f, _, _):
            conn = Mock()
            conn.alias = "default"
            conn.close_pool = Mock()
            f._db.connections.all = Mock(return_value=[conn])
            f._settings = DJSettings

            f._settings.DATABASES["default"] = {"OPTIONS": {}}
            f.close_database()
            conn.close.assert_called_once_with()
            conn.close_pool.assert_not_called()

            conn.reset_mock()
            f._settings.DATABASES["default"] = {"OPTIONS": {"pool": True}}
            f.close_database()
            conn.close.assert_called_once_with()
            conn.close_pool.assert_called_once_with()

            conn.reset_mock()
            f._settings.DATABASES["default"] = {"OPTIONS": {"pool": False}}
            f.close_database()
            conn.close.assert_called_once_with()
            conn.close_pool.assert_not_called()

    def test_close_database_conn_pool_thread_pool(self):
        class DJSettings:
            DATABASES = {}

        with self.fixup_context(self.app) as (f, _, _):
            conn = Mock()
            conn.alias = "default"
            conn.close_pool = Mock()
            f._db.connections.all = Mock(return_value=[conn])
            f._settings = DJSettings

            f._settings.DATABASES["default"] = {"OPTIONS": {"pool": True}}
            f.close_database()
            conn.close.assert_called_once_with()
            conn.close_pool.assert_called_once_with()

            conn.reset_mock()
            f.worker.pool_cls = ThreadTaskPool
            assert "prefork" not in ThreadTaskPool.__module__
            f.close_database()
            conn.close.assert_called_once_with()
            conn.close_pool.assert_not_called()

    def test_close_cache_raises_error(self):
        with self.fixup_context(self.app) as (f, _, _):
            f._cache.close_caches.side_effect = AttributeError
            f.close_cache()

    def test_close_cache(self):
        with self.fixup_context(self.app) as (f, _, _):
            f.close_cache()
            f._cache.close_caches.assert_called_with()

    @pytest.mark.patched_module('django', 'django.db', 'django.core',
                                'django.core.cache', 'django.conf',
                                'django.db.utils')
    def test_validate_models(self, patching, module):
        f = self.Fixup(self.app)
        f.django_setup = Mock(name='django.setup')
        patching.modules('django.core.checks')
        from django.core.checks import run_checks

        f.validate_models()
        f.django_setup.assert_called_with()
        run_checks.assert_called_with()

        # test --skip-checks flag
        f.django_setup.reset_mock()
        run_checks.reset_mock()

        patching.setenv('CELERY_SKIP_CHECKS', 'true')
        f.validate_models()
        f.django_setup.assert_called_with()
        run_checks.assert_not_called()

    def test_django_setup(self, patching):
        patching('celery.fixups.django.symbol_by_name')
        patching('celery.fixups.django.import_module')
        django, = patching.modules('django')
        f = self.Fixup(self.app)
        f.django_setup()
        django.setup.assert_called_with()

    def test__is_prefork(self):
        with self.fixup_context(self.app) as (f, _, _):
            f.worker.pool_cls = Mock(__module__='celery.concurrency.prefork')
            assert f._is_prefork()

            f.worker.pool_cls = "prefork"
            assert f._is_prefork()

            f.worker.pool_cls = Mock(__module__='celery.concurrency.thread')
            assert not f._is_prefork()

            f.worker = None
            assert not f._is_prefork()

    def test_no_recursive_worker_instantiation(self, patching):
        """Regression test: DjangoWorkerFixup must not create a WorkController in __init__.

        Historically, DjangoWorkerFixup.__init__ instantiated a WorkController when
        called with worker=None, which could cause recursive instantiation when
        invoked from worker lifecycle signals.

        This test verifies the fixed behavior:
        - DjangoWorkerFixup(app, worker=None) must not create a WorkController
        - It should instead leave self.worker unset and rely on on_worker_init
          to attach the actual worker instance later
        """
        from celery.worker import WorkController

        patching('celery.fixups.django.symbol_by_name')
        patching('celery.fixups.django.import_module')
        patching.modules('django', 'django.db', 'django.core.checks')

        # Track WorkController instantiations
        instantiation_count = {'count': 0}
        original_init = WorkController.__init__

        def tracking_init(self_worker, *args, **kwargs):
            instantiation_count['count'] += 1
            return original_init(self_worker, *args, **kwargs)

        with patch.object(WorkController, '__init__', tracking_init):
            # Creating DjangoWorkerFixup without a worker argument
            # should NOT create a WorkController instance
            DjangoWorkerFixup(self.app)

        # EXPECTED: 0 WorkController instances created
        assert instantiation_count['count'] == 0, (
            f"DjangoWorkerFixup(app) should NOT create a WorkController, "
            f"but {instantiation_count['count']} instance(s) were created. "
            f"This is the root cause of the recursion bug."
        )



import uuid

import pytest
from pytest_celery import RESULT_TIMEOUT, CeleryTestSetup

from celery.canvas import chain, chord, group, signature
from t.integration.conftest import get_redis_connection
from t.integration.tasks import ExpectedException, add, fail, identity, redis_echo


class test_signature:
    def test_sanity(self, celery_setup: CeleryTestSetup):
        sig = signature(identity, args=("test_signature",), queue=celery_setup.worker.worker_queue)
        assert sig.delay().get(timeout=RESULT_TIMEOUT) == "test_signature"


class test_group:
    def test_sanity(self, celery_setup: CeleryTestSetup):
        sig = group(
            group(add.si(1, 1), add.si(2, 2)),
            group([add.si(1, 1), add.si(2, 2)]),
            group(s for s in [add.si(1, 1), add.si(2, 2)]),
        )
        res = sig.apply_async(queue=celery_setup.worker.worker_queue)
        assert res.get(timeout=RESULT_TIMEOUT) == [2, 4, 2, 4, 2, 4]


class test_chain:
    def test_sanity(self, celery_setup: CeleryTestSetup):
        queue = celery_setup.worker.worker_queue
        sig = chain(
            identity.si("chain_task1").set(queue=queue),
            identity.si("chain_task2").set(queue=queue),
        ) | identity.si("test_chain").set(queue=queue)
        res = sig.apply_async()
        assert res.get(timeout=RESULT_TIMEOUT) == "test_chain"

    def test_chain_gets_last_task_id_with_failing_tasks_in_chain(self, celery_setup: CeleryTestSetup):
        """https://github.com/celery/celery/issues/8786"""
        queue = celery_setup.worker.worker_queue
        sig = chain(
            identity.si("start").set(queue=queue),
            group(
                identity.si("a").set(queue=queue),
                fail.si().set(queue=queue),
            ),
            identity.si("break").set(queue=queue),
            identity.si("end").set(queue=queue),
        )
        res = sig.apply_async()
        celery_setup.worker.assert_log_does_not_exist("ValueError: task_id must not be empty. Got None instead.")

        with pytest.raises(ExpectedException):
            res.get(timeout=RESULT_TIMEOUT)

    def test_upgrade_to_chord_inside_chains(self, celery_setup: CeleryTestSetup):
        redis_key = str(uuid.uuid4())
        queue = celery_setup.worker.worker_queue
        group1 = group(redis_echo.si("a", redis_key), redis_echo.si("a", redis_key))
        group2 = group(redis_echo.si("a", redis_key), redis_echo.si("a", redis_key))
        chord1 = group1 | group2
        chain1 = chain(chord1, (redis_echo.si("a", redis_key) | redis_echo.si("b", redis_key).set(queue=queue)))
        chain1.apply_async(queue=queue).get(timeout=RESULT_TIMEOUT)
        redis_connection = get_redis_connection()
        actual = redis_connection.lrange(redis_key, 0, -1)
        assert actual.count(b"a") == 5
        assert actual.count(b"b") == 1
        redis_connection.delete(redis_key)


class test_chord:
    def test_sanity(self, celery_setup: CeleryTestSetup):
        upgraded_chord = signature(
            group(
                identity.si("header_task1"),
                identity.si("header_task2"),
            )
            | identity.si("body_task"),
            queue=celery_setup.worker.worker_queue,
        )

        sig = group(
            [
                upgraded_chord,
                chord(
                    group(
                        identity.si("header_task3"),
                        identity.si("header_task4"),
                    ),
                    identity.si("body_task"),
                ),
                chord(
                    (
                        sig
                        for sig in [
                            identity.si("header_task5"),
                            identity.si("header_task6"),
                        ]
                    ),
                    identity.si("body_task"),
                ),
            ]
        )
        res = sig.apply_async(queue=celery_setup.worker.worker_queue)
        assert res.get(timeout=RESULT_TIMEOUT) == ["body_task"] * 3

    @pytest.mark.parametrize(
        "input_body",
        [
            (lambda queue: add.si(9, 7).set(queue=queue)),
            (
                lambda queue: chain(
                    add.si(9, 7).set(queue=queue),
                    add.si(5, 7).set(queue=queue),
                )
            ),
            (
                lambda queue: group(
                    [
                        add.si(9, 7).set(queue=queue),
                        add.si(5, 7).set(queue=queue),
                    ]
                )
            ),
            (
                lambda queue: chord(
                    group(
                        [
                            add.si(1, 1).set(queue=queue),
                            add.si(2, 2).set(queue=queue),
                        ]
                    ),
                    add.si(10, 10).set(queue=queue),
                )
            ),
        ],
        ids=[
            "body is a single_task",
            "body is a chain",
            "body is a group",
            "body is a chord",
        ],
    )
    def test_chord_error_propagation_with_different_body_types(
        self, celery_setup: CeleryTestSetup, input_body
    ) -> None:
        """Reproduce issue #9773 with different chord body types.

        This test verifies that the "task_id must not be empty" error is fixed
        regardless of the chord body type. The issue occurs when:
        1. A chord has a group with both succeeding and failing tasks
        2. The chord body can be any signature type (single task, chain, group, chord)
        3. When the group task fails, error propagation should work correctly

        Args:
            input_body (callable): A callable that returns a Celery signature for the chord body.
        """
        queue = celery_setup.worker.worker_queue

        # Create the failing group header (same for all tests)
        failing_group = group(
            [
                add.si(15, 7).set(queue=queue),
                # failing task
                fail.si().set(queue=queue),
            ]
        )

        # Create the chord
        test_chord = chord(failing_group, input_body(queue))

        result = test_chord.apply_async()

        # The worker should not log the "task_id must not be empty" error
        celery_setup.worker.assert_log_does_not_exist(
            "ValueError: task_id must not be empty. Got None instead."
        )

        # The chord should fail with the expected exception from the failing task
        with pytest.raises(ExpectedException):
            result.get(timeout=RESULT_TIMEOUT)


class test_complex_workflow:
    def test_pending_tasks_released_on_forget(self, celery_setup: CeleryTestSetup):
        sig = add.si(1, 1) | group(
            add.s(1) | group(add.si(1, 1), add.si(2, 2)) | add.si(2, 2),
            add.s(1) | group(add.si(1, 1), add.si(2, 2)) | add.si(2, 2)
        ) | add.si(1, 1)
        res = sig.apply_async(queue=celery_setup.worker.worker_queue)
        assert not all(len(mapping) == 0 for mapping in res.backend._pending_results)
        res.forget()
        assert all(len(mapping) == 0 for mapping in res.backend._pending_results)



import inspect
import sys
import typing

import pytest
from pydantic import BaseModel

from celery.utils.annotations import annotation_issubclass, get_optional_arg, is_none_type


@pytest.mark.parametrize(
    'value,expected',
    ((3, False), ('x', False), (int, False), (type(None), True)),
)
def test_is_none_type(value: typing.Any, expected: bool) -> None:
    assert is_none_type(value) is expected


def test_is_none_type_with_optional_annotations() -> None:
    annotation = typing.Optional[int]
    int_type, none_type = typing.get_args(annotation)
    assert int_type == int  # just to make sure that order is correct
    assert is_none_type(int_type) is False
    assert is_none_type(none_type) is True


def test_get_optional_arg() -> None:
    def func(
        arg: int,
        optional: typing.Optional[int],
        optional2: typing.Union[int, None],
        optional3: typing.Union[None, int],
        not_optional1: typing.Union[str, int],
        not_optional2: typing.Union[str, int, bool],
    ) -> None:
        pass

    parameters = inspect.signature(func).parameters

    assert get_optional_arg(parameters['arg'].annotation) is None
    assert get_optional_arg(parameters['optional'].annotation) is int
    assert get_optional_arg(parameters['optional2'].annotation) is int
    assert get_optional_arg(parameters['optional3'].annotation) is int
    assert get_optional_arg(parameters['not_optional1'].annotation) is None
    assert get_optional_arg(parameters['not_optional2'].annotation) is None


@pytest.mark.skipif(sys.version_info < (3, 10), reason="Notation is only supported in Python 3.10 or newer.")
def test_get_optional_arg_with_pipe_notation() -> None:
    def func(optional: int | None, optional2: None | int) -> None:
        pass

    parameters = inspect.signature(func).parameters

    assert get_optional_arg(parameters['optional'].annotation) is int
    assert get_optional_arg(parameters['optional2'].annotation) is int


def test_annotation_issubclass() -> None:
    def func(
        int_arg: int,
        base_model: BaseModel,
        list_arg: list,  # type: ignore[type-arg]  # what we test
        dict_arg: dict,  # type: ignore[type-arg]  # what we test
        list_typing_arg: typing.List,  # type: ignore[type-arg]  # what we test
        dict_typing_arg: typing.Dict,  # type: ignore[type-arg]  # what we test
        list_typing_generic_arg: typing.List[str],
        dict_typing_generic_arg: typing.Dict[str, str],
    ) -> None:
        pass

    parameters = inspect.signature(func).parameters
    assert annotation_issubclass(parameters['int_arg'].annotation, int) is True
    assert annotation_issubclass(parameters['base_model'].annotation, BaseModel) is True
    assert annotation_issubclass(parameters['list_arg'].annotation, list) is True
    assert annotation_issubclass(parameters['dict_arg'].annotation, dict) is True

    # Here the annotation is simply not a class, so function must return False
    assert annotation_issubclass(parameters['list_typing_arg'].annotation, BaseModel) is False
    assert annotation_issubclass(parameters['dict_typing_arg'].annotation, BaseModel) is False
    assert annotation_issubclass(parameters['list_typing_generic_arg'].annotation, BaseModel) is False
    assert annotation_issubclass(parameters['dict_typing_generic_arg'].annotation, BaseModel) is False


@pytest.mark.skipif(sys.version_info < (3, 9), reason="Notation is only supported in Python 3.9 or newer.")
def test_annotation_issubclass_with_generic_classes() -> None:
    def func(list_arg: list[str], dict_arg: dict[str, str]) -> None:
        pass

    parameters = inspect.signature(func).parameters
    assert annotation_issubclass(parameters['list_arg'].annotation, list) is False
    assert annotation_issubclass(parameters['dict_arg'].annotation, dict) is False

    # issubclass() behaves differently with BaseModel (and maybe other classes?).
    assert annotation_issubclass(parameters['list_arg'].annotation, BaseModel) is False
    assert annotation_issubclass(parameters['dict_arg'].annotation, BaseModel) is False
